import re

import pytest

from faker.providers.lorem.az_AZ import Provider as AzAzLoremProvider
from faker.providers.lorem.bn_BD import Provider as BnBdLoremProvider
from faker.providers.lorem.cs_CZ import Provider as CsCzLoremProvider
from faker.providers.lorem.de_AT import Provider as DeAtLoremProvider
from faker.providers.lorem.de_DE import Provider as DeDeLoremProvider
from faker.providers.lorem.en_US import Provider as EnUsLoremProvider
from faker.providers.lorem.fa_IR import Provider as FaIrLoremProvider
from faker.providers.lorem.nl_BE import Provider as NlBeLoremProvider
from faker.providers.lorem.uk_UA import Provider as UkUaLoremProvider


class TestLoremProvider:
    """Test lorem provider methods"""

    custom_word_list = ["danish", "cheesecake", "sugar", "lollipop", "wafer", "gummies", "jelly", "pie"]

    def test_word_with_defaults(self, faker, num_samples):
        for _ in range(num_samples):
            word = faker.word()
            assert isinstance(word, str)

    def test_word_with_custom_list(self, faker, num_samples):
        for _ in range(num_samples):
            word = faker.word(ext_word_list=self.custom_word_list)
            assert isinstance(word, str)
            assert word in self.custom_word_list

    def test_words_with_zero_nb(self, faker, num_samples):
        for _ in range(num_samples):
            assert faker.words(0) == []

    def test_words_with_defaults(self, faker, num_samples):
        num_words = 5
        for _ in range(num_samples):
            words = faker.words(num_words)
            assert isinstance(words, list)
            assert len(words) == 5
            assert all(isinstance(word, str) for word in words)

    def test_words_with_custom_word_list(self, faker, num_samples):
        num_words = 5
        for _ in range(num_samples):
            words = faker.words(num_words, ext_word_list=self.custom_word_list)
            assert isinstance(words, list)
            assert len(words) == 5
            assert all(isinstance(word, str) and word in self.custom_word_list for word in words)

    def test_words_with_unique_sampling(self, faker, num_samples):
        num_words = 5
        for _ in range(num_samples):
            words = faker.words(num_words, ext_word_list=self.custom_word_list, unique=True)
            assert isinstance(words, list)
            assert len(words) == 5

            checked_words = []
            for word in words:
                assert isinstance(word, str)
                assert word in self.custom_word_list
                assert word not in checked_words
                checked_words.append(word)

    def test_sentence_no_words(self, faker, num_samples):
        for _ in range(num_samples):
            assert faker.sentence(0) == ""

    def test_sentence_with_inexact_word_count(self, faker, num_samples):
        num_words = 10
        for _ in range(num_samples):
            sentence = faker.sentence(nb_words=num_words)
            words = sentence.lower().replace(".", "").split()
            assert int(num_words * 0.6) <= len(words) <= int(num_words * 1.4)

    def test_sentence_with_exact_word_count(self, faker, num_samples):
        num_words = 10
        for _ in range(num_samples):
            sentence = faker.sentence(nb_words=num_words, variable_nb_words=False)
            words = sentence.lower().replace(".", "").split()
            assert len(words) == num_words

    def test_sentence_with_custom_word_list(self, faker, num_samples):
        for _ in range(num_samples):
            sentence = faker.sentence(ext_word_list=self.custom_word_list)
            words = sentence.lower().replace(".", "").split()
            assert all(isinstance(word, str) and word in self.custom_word_list for word in words)

    def test_sentences(self, faker, num_samples):
        num_sentences = 5
        for _ in range(num_samples):
            sentences = faker.sentences(nb=num_sentences, ext_word_list=self.custom_word_list)
            assert len(sentences) == num_sentences
            for sentence in sentences:
                words = sentence.lower().replace(".", "").split()
                assert all(isinstance(word, str) and word in self.custom_word_list for word in words)

    def test_sentence_single_word(self, faker):
        word = faker.sentence(1)
        assert str.isupper(word[0])

    def test_paragraph_no_sentences(self, faker, num_samples):
        for _ in range(num_samples):
            assert faker.paragraph(0) == ""

    def test_paragraph_with_inexact_sentence_count(self, faker, num_samples):
        num_sentences = 10
        for _ in range(num_samples):
            paragraph = faker.paragraph(nb_sentences=num_sentences)
            sentences = paragraph.split(". ")
            assert int(num_sentences * 0.6) <= len(sentences) <= int(num_sentences * 1.4)

    def test_paragraph_with_exact_sentence_count(self, faker, num_samples):
        num_sentences = 10
        for _ in range(num_samples):
            paragraph = faker.paragraph(nb_sentences=num_sentences, variable_nb_sentences=False)
            sentences = paragraph.split(". ")
            assert len(sentences) == num_sentences

    def test_paragraph_with_custom_word_list(self, faker, num_samples):
        for _ in range(num_samples):
            paragraph = faker.paragraph(ext_word_list=self.custom_word_list)
            words = paragraph.lower().replace(".", "").split()
            assert all(isinstance(word, str) and word in self.custom_word_list for word in words)

    def test_paragraphs(self, faker, num_samples):
        num_paragraphs = 5
        for _ in range(num_samples):
            paragraphs = faker.paragraphs(nb=num_paragraphs, ext_word_list=self.custom_word_list)
            assert len(paragraphs) == num_paragraphs
            for paragraph in paragraphs:
                words = paragraph.lower().replace(".", "").split()
                assert all(isinstance(word, str) and word in self.custom_word_list for word in words)

    def test_text_with_less_than_four_characters(self, faker, num_samples):
        for _ in range(num_samples):
            with pytest.raises(ValueError):
                faker.text(max_nb_chars=4)

    @pytest.mark.parametrize(
        "num_chars",
        [10, 50, 150, 10000],
        ids=["max_nb_chars < 25", "25 <= max_nb_chars < 100", "max_nb_chars >= 100", "max_nb_chars >> 100"],
    )
    def test_text_with_valid_character_count(self, faker, num_samples, num_chars):
        for _ in range(num_samples):
            text = faker.text(max_nb_chars=num_chars)
            assert len(text) <= num_chars

    def test_text_with_custom_word_list(self, faker, num_samples):
        for _ in range(num_samples):
            text = faker.text(ext_word_list=self.custom_word_list)
            words = re.sub(r"[.\n]+", " ", text.lower()).split()
            assert all(word in self.custom_word_list for word in words)

    def test_texts(self, faker, num_samples):
        num_texts = 5
        num_chars = 25
        for _ in range(num_samples):
            texts = faker.texts(max_nb_chars=num_chars, nb_texts=num_texts, ext_word_list=self.custom_word_list)
            assert len(texts) == num_texts
            for text in texts:
                assert len(text) <= num_chars
                words = re.sub(r"[.\n]+", " ", text.lower()).split()
                assert all(word in self.custom_word_list for word in words)

    @pytest.mark.parametrize(
        "nb,part_of_speech", [(10, "verb"), (18, "adverb"), (11, "noun")], ids=["verb", "adverb", "noun"]
    )
    def test_words_part_of_speech(self, faker, nb, part_of_speech):
        words = faker.words(nb=nb, part_of_speech=part_of_speech)
        assert (word in EnUsLoremProvider.parts_of_speech[part_of_speech] for word in words)

    @pytest.mark.parametrize("nb,part_of_speech", [(5, "abcdefg")], ids=["invalid part of speech"])
    def test_words_invalid_part_of_speech(self, faker, nb, part_of_speech):
        with pytest.raises(ValueError) as exc_info:
            faker.words(nb=nb, part_of_speech=part_of_speech)

        assert exc_info.type is ValueError
        assert exc_info.value.args[0] == f"{part_of_speech} is not recognized as a part of speech."

    @pytest.mark.parametrize(
        "nb,part_of_speech",
        [(3, "adverb"), (5, "verb"), (4, "abcdefgh")],
        ids=["ignore adverb", "ignore verb", "ignore invalid part of speech"],
    )
    def test_words_part_of_speech_ignored(self, faker, nb, part_of_speech):
        words = faker.words(nb=nb, part_of_speech=part_of_speech, ext_word_list=self.custom_word_list)
        assert len(words) == nb
        assert all(word in self.custom_word_list for word in words)


class TestCsCz:
    """Test cs_CZ lorem provider"""

    word_list = [word.lower() for word in CsCzLoremProvider.word_list]

    def test_paragraph(self, faker, num_samples):
        num_sentences = 10
        for _ in range(num_samples):
            paragraph = faker.paragraph(nb_sentences=num_sentences)
            assert isinstance(paragraph, str)
            words = paragraph.replace(".", "").split()
            assert all(word.lower() in self.word_list for word in words)

    def test_paragraphs(self, faker, num_samples):
        num_paragraphs = 5
        for _ in range(num_samples):
            paragraphs = faker.paragraphs(nb=num_paragraphs)
            for paragraph in paragraphs:
                assert isinstance(paragraph, str)
                words = paragraph.replace(".", "").split()
                assert all(word.lower() in self.word_list for word in words)

    def test_sentence(self, faker, num_samples):
        num_words = 10
        for _ in range(num_samples):
            sentence = faker.sentence(nb_words=num_words)
            assert isinstance(sentence, str)
            words = sentence.replace(".", "").split()
            assert all(word.lower() in self.word_list for word in words)

    def test_sentences(self, faker, num_samples):
        num_sentences = 5
        for _ in range(num_samples):
            sentences = faker.sentences(nb=num_sentences)
            for sentence in sentences:
                assert isinstance(sentence, str)
                words = sentence.replace(".", "").split()
                assert all(word.lower() in self.word_list for word in words)

    def test_text(self, faker, num_samples):
        num_chars = 25
        for _ in range(num_samples):
            text = faker.text(max_nb_chars=num_chars)
            assert isinstance(text, str)
            words = re.sub(r"[.\n]+", " ", text).split()
            assert all(word.lower() in self.word_list for word in words)

    def test_texts(self, faker, num_samples):
        num_texts = 5
        num_chars = 25
        for _ in range(num_samples):
            texts = faker.texts(max_nb_chars=num_chars, nb_texts=num_texts)
            for text in texts:
                assert isinstance(text, str)
                words = re.sub(r"[.\n]+", " ", text).split()
                assert all(word.lower() in self.word_list for word in words)

    def test_word(self, faker, num_samples):
        for _ in range(num_samples):
            word = faker.word()
            assert isinstance(word, str) and word in CsCzLoremProvider.word_list

    def test_words(self, faker, num_samples):
        num_words = 5
        for _ in range(num_samples):
            words = faker.words(num_words)
            assert all(isinstance(word, str) and word in CsCzLoremProvider.word_list for word in words)


class TestAzAz:
    """Test az_AZ lorem provider"""

    word_list = [word.lower() for word in AzAzLoremProvider.word_list]

    def test_paragraph(self, faker, num_samples):
        num_sentences = 10
        for _ in range(num_samples):
            paragraph = faker.paragraph(nb_sentences=num_sentences)
            assert isinstance(paragraph, str)
            words = paragraph.replace(".", "").split()
            assert all(word.lower() in self.word_list for word in words)

    def test_paragraphs(self, faker, num_samples):
        num_paragraphs = 5
        for _ in range(num_samples):
            paragraphs = faker.paragraphs(nb=num_paragraphs)
            for paragraph in paragraphs:
                assert isinstance(paragraph, str)
                words = paragraph.replace(".", "").split()
                assert all(word.lower() in self.word_list for word in words)

    def test_sentence(self, faker, num_samples):
        num_words = 10
        for _ in range(num_samples):
            sentence = faker.sentence(nb_words=num_words)
            assert isinstance(sentence, str)
            words = sentence.replace(".", "").split()
            assert all(word.lower() in self.word_list for word in words)

    def test_sentences(self, faker, num_samples):
        num_sentences = 5
        for _ in range(num_samples):
            sentences = faker.sentences(nb=num_sentences)
            for sentence in sentences:
                assert isinstance(sentence, str)
                words = sentence.replace(".", "").split()
                assert all(word.lower() in self.word_list for word in words)

    def test_text(self, faker, num_samples):
        num_chars = 25
        for _ in range(num_samples):
            text = faker.text(max_nb_chars=num_chars)
            assert isinstance(text, str)
            words = re.sub(r"[.\n]+", " ", text).split()
            assert all(word.lower() in self.word_list for word in words)

    def test_texts(self, faker, num_samples):
        num_texts = 5
        num_chars = 25
        for _ in range(num_samples):
            texts = faker.texts(max_nb_chars=num_chars, nb_texts=num_texts)
            for text in texts:
                assert isinstance(text, str)
                words = re.sub(r"[.\n]+", " ", text).split()
                assert all(word.lower() in self.word_list for word in words)

    def test_word(self, faker, num_samples):
        for _ in range(num_samples):
            word = faker.word()
            assert isinstance(word, str) and word in AzAzLoremProvider.word_list

    def test_words(self, faker, num_samples):
        num_words = 5
        for _ in range(num_samples):
            words = faker.words(num_words)
            assert all(isinstance(word, str) and word in AzAzLoremProvider.word_list for word in words)


class TestFaIr:
    """Test fa_IR lorem provider"""

    word_list = [word.lower() for word in FaIrLoremProvider.word_list]

    def test_paragraph(self, faker, num_samples):
        num_sentences = 10
        for _ in range(num_samples):
            paragraph = faker.paragraph(nb_sentences=num_sentences)
            assert isinstance(paragraph, str)
            words = paragraph.replace(".", "").split()
            assert all(word.lower() in self.word_list for word in words)

    def test_paragraphs(self, faker, num_samples):
        num_paragraphs = 5
        for _ in range(num_samples):
            paragraphs = faker.paragraphs(nb=num_paragraphs)
            for paragraph in paragraphs:
                assert isinstance(paragraph, str)
                words = paragraph.replace(".", "").split()
                assert all(word.lower() in self.word_list for word in words)

    def test_sentence(self, faker, num_samples):
        num_words = 10
        for _ in range(num_samples):
            sentence = faker.sentence(nb_words=num_words)
            assert isinstance(sentence, str)
            words = sentence.replace(".", "").split()
            assert all(word.lower() in self.word_list for word in words)

    def test_sentences(self, faker, num_samples):
        num_sentences = 5
        for _ in range(num_samples):
            sentences = faker.sentences(nb=num_sentences)
            for sentence in sentences:
                assert isinstance(sentence, str)
                words = sentence.replace(".", "").split()
                assert all(word.lower() in self.word_list for word in words)

    def test_text(self, faker, num_samples):
        num_chars = 25
        for _ in range(num_samples):
            text = faker.text(max_nb_chars=num_chars)
            assert isinstance(text, str)
            words = re.sub(r"[.\n]+", " ", text).split()
            assert all(word.lower() in self.word_list for word in words)

    def test_texts(self, faker, num_samples):
        num_texts = 5
        num_chars = 25
        for _ in range(num_samples):
            texts = faker.texts(max_nb_chars=num_chars, nb_texts=num_texts)
            for text in texts:
                assert isinstance(text, str)
                words = re.sub(r"[.\n]+", " ", text).split()
                assert all(word.lower() in self.word_list for word in words)

    def test_word(self, faker, num_samples):
        for _ in range(num_samples):
            word = faker.word()
            assert isinstance(word, str) and word in FaIrLoremProvider.word_list

    def test_words(self, faker, num_samples):
        num_words = 5
        for _ in range(num_samples):
            words = faker.words(num_words)
            assert all(isinstance(word, str) and word in FaIrLoremProvider.word_list for word in words)


class TestBnBd:
    """Test bn_BD lorem provider"""

    word_list = BnBdLoremProvider.word_list

    def test_paragraph(self, faker, num_samples):
        num_sentences = 10
        for _ in range(num_samples):
            paragraph = faker.paragraph(nb_sentences=num_sentences)
            assert isinstance(paragraph, str)
            words = paragraph.replace("।", "").split()
            assert all(word in self.word_list for word in words)

    def test_paragraphs(self, faker, num_samples):
        num_paragraphs = 5
        for _ in range(num_samples):
            paragraphs = faker.paragraphs(nb=num_paragraphs)
            for paragraph in paragraphs:
                assert isinstance(paragraph, str)
                words = paragraph.replace("।", "").split()
                assert all(word in self.word_list for word in words)

    def test_sentence(self, faker, num_samples):
        num_words = 10
        for _ in range(num_samples):
            sentence = faker.sentence(nb_words=num_words)
            assert isinstance(sentence, str)
            words = sentence.replace("।", "").split()
            assert all(word in self.word_list for word in words)

    def test_sentences(self, faker, num_samples):
        num_sentences = 5
        for _ in range(num_samples):
            sentences = faker.sentences(nb=num_sentences)
            for sentence in sentences:
                assert isinstance(sentence, str)
                words = sentence.replace("।", "").split()
                assert all(word in self.word_list for word in words)


class TestDeDe:
    """Test ```de_DE``` lorem provider"""

    word_list = [word.lower() for word in DeDeLoremProvider.word_list]

    def test_paragraph(self, faker, num_samples):
        num_sentences = 10
        for _ in range(num_samples):
            paragraph = faker.paragraph(nb_sentences=num_sentences)
            assert isinstance(paragraph, str)
            words = paragraph.replace(".", "").split()
            assert all(word.lower() in self.word_list for word in words)

    def test_paragraphs(self, faker, num_samples):
        num_paragraphs = 5
        for _ in range(num_samples):
            paragraphs = faker.paragraphs(nb=num_paragraphs)
            for paragraph in paragraphs:
                assert isinstance(paragraph, str)
                words = paragraph.replace(".", "").split()
                assert all(word.lower() in self.word_list for word in words)

    def test_sentence(self, faker, num_samples):
        num_words = 10
        for _ in range(num_samples):
            sentence = faker.sentence(nb_words=num_words)
            assert isinstance(sentence, str)
            words = sentence.replace(".", "").split()
            assert all(word.lower() in self.word_list for word in words)

    def test_sentences(self, faker, num_samples):
        num_sentences = 5
        for _ in range(num_samples):
            sentences = faker.sentences(nb=num_sentences)
            for sentence in sentences:
                assert isinstance(sentence, str)
                words = sentence.replace(".", "").split()
                assert all(word.lower() in self.word_list for word in words)

    def test_text(self, faker, num_samples):
        num_chars = 25
        for _ in range(num_samples):
            text = faker.text(max_nb_chars=num_chars)
            assert isinstance(text, str)
            words = re.sub(r"[.\n]+", " ", text).split()
            assert all(word.lower() in self.word_list for word in words)

    def test_texts(self, faker, num_samples):
        num_texts = 5
        num_chars = 25
        for _ in range(num_samples):
            texts = faker.texts(max_nb_chars=num_chars, nb_texts=num_texts)
            for text in texts:
                assert isinstance(text, str)
                words = re.sub(r"[.\n]+", " ", text).split()
                assert all(word.lower() in self.word_list for word in words)

    def test_word(self, faker, num_samples):
        for _ in range(num_samples):
            word = faker.word()
            assert isinstance(word, str) and word in DeDeLoremProvider.word_list

    def test_words(self, faker, num_samples):
        num_words = 5
        for _ in range(num_samples):
            words = faker.words(num_words)
            assert all(isinstance(word, str) and word in DeDeLoremProvider.word_list for word in words)


class TestDeAt:
    """Test ```de_AT``` lorem provider"""

    word_list = [word.lower() for word in DeAtLoremProvider.word_list]

    def test_paragraph(self, faker, num_samples):
        num_sentences = 10
        for _ in range(num_samples):
            paragraph = faker.paragraph(nb_sentences=num_sentences)
            assert isinstance(paragraph, str)
            words = paragraph.replace(".", "").split()
            assert all(word.lower() in self.word_list for word in words)

    def test_paragraphs(self, faker, num_samples):
        num_paragraphs = 5
        for _ in range(num_samples):
            paragraphs = faker.paragraphs(nb=num_paragraphs)
            for paragraph in paragraphs:
                assert isinstance(paragraph, str)
                words = paragraph.replace(".", "").split()
                assert all(word.lower() in self.word_list for word in words)

    def test_sentence(self, faker, num_samples):
        num_words = 10
        for _ in range(num_samples):
            sentence = faker.sentence(nb_words=num_words)
            assert isinstance(sentence, str)
            words = sentence.replace(".", "").split()
            assert all(word.lower() in self.word_list for word in words)

    def test_sentences(self, faker, num_samples):
        num_sentences = 5
        for _ in range(num_samples):
            sentences = faker.sentences(nb=num_sentences)
            for sentence in sentences:
                assert isinstance(sentence, str)
                words = sentence.replace(".", "").split()
                assert all(word.lower() in self.word_list for word in words)

    def test_text(self, faker, num_samples):
        num_chars = 25
        for _ in range(num_samples):
            text = faker.text(max_nb_chars=num_chars)
            assert isinstance(text, str)
            words = re.sub(r"[.\n]+", " ", text).split()
            assert all(word.lower() in self.word_list for word in words)

    def test_texts(self, faker, num_samples):
        num_texts = 5
        num_chars = 25
        for _ in range(num_samples):
            texts = faker.texts(max_nb_chars=num_chars, nb_texts=num_texts)
            for text in texts:
                assert isinstance(text, str)
                words = re.sub(r"[.\n]+", " ", text).split()
                assert all(word.lower() in self.word_list for word in words)

    def test_word(self, faker, num_samples):
        for _ in range(num_samples):
            word = faker.word()
            assert isinstance(word, str) and word in DeAtLoremProvider.word_list

    def test_words(self, faker, num_samples):
        num_words = 5
        for _ in range(num_samples):
            words = faker.words(num_words)
            assert all(isinstance(word, str) and word in DeAtLoremProvider.word_list for word in words)


class TestNlBe:
    """Test ```nl_BE``` lorem provider

    Copied from the TestDeDe class, but with the word_list from the NlBeLoremProvider.
    """

    word_list = [word.lower() for word in NlBeLoremProvider.word_list]

    def test_paragraph(self, faker, num_samples):
        num_sentences = 10
        for _ in range(num_samples):
            paragraph = faker.paragraph(nb_sentences=num_sentences)
            assert isinstance(paragraph, str)
            words = paragraph.replace(".", "").split()
            assert all(word.lower() in self.word_list for word in words)

    def test_paragraphs(self, faker, num_samples):
        num_paragraphs = 5
        for _ in range(num_samples):
            paragraphs = faker.paragraphs(nb=num_paragraphs)
            for paragraph in paragraphs:
                assert isinstance(paragraph, str)
                words = paragraph.replace(".", "").split()
                assert all(word.lower() in self.word_list for word in words)

    def test_sentence(self, faker, num_samples):
        num_words = 10
        for _ in range(num_samples):
            sentence = faker.sentence(nb_words=num_words)
            assert isinstance(sentence, str)
            words = sentence.replace(".", "").split()
            assert all(word.lower() in self.word_list for word in words)

    def test_sentences(self, faker, num_samples):
        num_sentences = 5
        for _ in range(num_samples):
            sentences = faker.sentences(nb=num_sentences)
            for sentence in sentences:
                assert isinstance(sentence, str)
                words = sentence.replace(".", "").split()
                assert all(word.lower() in self.word_list for word in words)

    def test_text(self, faker, num_samples):
        num_chars = 25
        for _ in range(num_samples):
            text = faker.text(max_nb_chars=num_chars)
            assert isinstance(text, str)
            words = re.sub(r"[.\n]+", " ", text).split()
            assert all(word.lower() in self.word_list for word in words)

    def test_texts(self, faker, num_samples):
        num_texts = 5
        num_chars = 25
        for _ in range(num_samples):
            texts = faker.texts(max_nb_chars=num_chars, nb_texts=num_texts)
            for text in texts:
                assert isinstance(text, str)
                words = re.sub(r"[.\n]+", " ", text).split()
                assert all(word.lower() in self.word_list for word in words)

    def test_word(self, faker, num_samples):
        for _ in range(num_samples):
            word = faker.word()
            assert isinstance(word, str) and word in NlBeLoremProvider.word_list

    def test_words(self, faker, num_samples):
        num_words = 5
        for _ in range(num_samples):
            words = faker.words(num_words)
            assert all(isinstance(word, str) and word in NlBeLoremProvider.word_list for word in words)


class TestUkUa:
    """Test uk_UA lorem provider"""

    word_list = [word.lower() for word in UkUaLoremProvider.word_list]

    def test_paragraph(self, faker, num_samples):
        num_sentences = 10
        for _ in range(num_samples):
            paragraph = faker.paragraph(nb_sentences=num_sentences)
            assert isinstance(paragraph, str)
            words = paragraph.replace(".", "").split()
            assert all(word.lower() in self.word_list for word in words)

    def test_paragraphs(self, faker, num_samples):
        num_paragraphs = 5
        for _ in range(num_samples):
            paragraphs = faker.paragraphs(nb=num_paragraphs)
            for paragraph in paragraphs:
                assert isinstance(paragraph, str)
                words = paragraph.replace(".", "").split()
                assert all(word.lower() in self.word_list for word in words)

    def test_sentence(self, faker, num_samples):
        num_words = 10
        for _ in range(num_samples):
            sentence = faker.sentence(nb_words=num_words)
            assert isinstance(sentence, str)
            words = sentence.replace(".", "").split()
            assert all(word.lower() in self.word_list for word in words)

    def test_sentences(self, faker, num_samples):
        num_sentences = 5
        for _ in range(num_samples):
            sentences = faker.sentences(nb=num_sentences)
            for sentence in sentences:
                assert isinstance(sentence, str)
                words = sentence.replace(".", "").split()
                assert all(word.lower() in self.word_list for word in words)

    def test_text(self, faker, num_samples):
        num_chars = 25
        for _ in range(num_samples):
            text = faker.text(max_nb_chars=num_chars)
            assert isinstance(text, str)
            words = re.sub(r"[.\n]+", " ", text).split()
            assert all(word.lower() in self.word_list for word in words)

    def test_texts(self, faker, num_samples):
        num_texts = 5
        num_chars = 25
        for _ in range(num_samples):
            texts = faker.texts(max_nb_chars=num_chars, nb_texts=num_texts)
            for text in texts:
                assert isinstance(text, str)
                words = re.sub(r"[.\n]+", " ", text).split()
                assert all(word.lower() in self.word_list for word in words)

    def test_word(self, faker, num_samples):
        for _ in range(num_samples):
            word = faker.word()
            assert isinstance(word, str) and word in UkUaLoremProvider.word_list

    def test_words(self, faker, num_samples):
        num_words = 5
        for _ in range(num_samples):
            words = faker.words(num_words)
            assert all(isinstance(word, str) and word in UkUaLoremProvider.word_list for word in words)



from .. import Provider as InternetProvider


class Provider(InternetProvider):
    safe_email_tlds = ("org", "com", "net", "ch")
    free_email_domains = (
        "gmail.com",
        "hotmail.fr",
        "yahoo.fr",
        "bluewin.ch",
        "romandie.com",
        "hispeed.ch",
        "sunrise.ch",
        "vtxnet.ch",
    )
    tlds = ("com", "com", "com", "net", "org", "ch", "ch", "ch")

    replacements = (
        ("ä", "ae"),
        ("à", "a"),
        ("â", "a"),
        ("ç", "c"),
        ("é", "e"),
        ("è", "e"),
        ("ê", "e"),
        ("ë", "e"),
        ("ï", "i"),
        ("î", "i"),
        ("ö", "oe"),
        ("ô", "o"),
        ("ü", "ue"),
        ("ù", "u"),
        ("ü", "u"),
        ("ß", "ss"),
    )



from .. import Provider as CompanyProvider


class Provider(CompanyProvider):
    formats = (
        "{{last_name}} {{company_suffix}}",
        "{{last_name}} {{last_name}} {{company_suffix}}",
        "{{last_name}} és {{last_name}} {{company_suffix}}",
        "{{last_name}} és társa {{company_suffix}}",
    )

    company_suffixes = ("Kft.", "Kht.", "Zrt.", "Bt.", "Nyrt.", "Kkt.")

    def company_suffix(self) -> str:
        return self.random_element(self.company_suffixes)



import re

from typing import Optional

from faker.providers import ElementsType

from ..en import Provider as AddressProvider


class Provider(AddressProvider):
    #  Source: https://www.canadapost.ca/tools/pg/manual/PGaddress-e.asp#1449294
    #
    #  'W' and 'Z' are valid in non-initial position (easily verified in the
    #  wild), but online official documentation is hard to find, so just ignore
    #  them for now.
    postal_code_letters = (
        "A",
        "B",
        "C",
        "E",
        "G",
        "H",
        "J",
        "K",
        "L",
        "M",
        "N",
        "P",
        "R",
        "S",
        "T",
        "V",
        "X",
        "Y",
    )

    city_prefixes: ElementsType[str] = ("North", "East", "West", "South", "New", "Lake", "Port")

    city_suffixes: ElementsType[str] = (
        "town",
        "ton",
        "land",
        "ville",
        "berg",
        "burgh",
        "borough",
        "bury",
        "view",
        "port",
        "mouth",
        "stad",
        "furt",
        "chester",
        "mouth",
        "fort",
        "haven",
        "side",
        "shire",
    )

    building_number_formats = ("#####", "####", "###")

    street_suffixes = (
        "Alley",
        "Avenue",
        "Branch",
        "Bridge",
        "Brook",
        "Brooks",
        "Burg",
        "Burgs",
        "Bypass",
        "Camp",
        "Canyon",
        "Cape",
        "Causeway",
        "Center",
        "Centers",
        "Circle",
        "Circles",
        "Cliff",
        "Cliffs",
        "Club",
        "Common",
        "Corner",
        "Corners",
        "Course",
        "Court",
        "Courts",
        "Cove",
        "Coves",
        "Creek",
        "Crescent",
        "Crest",
        "Crossing",
        "Crossroad",
        "Curve",
        "Dale",
        "Dam",
        "Divide",
        "Drive",
        "Drive",
        "Drives",
        "Estate",
        "Estates",
        "Expressway",
        "Extension",
        "Extensions",
        "Fall",
        "Falls",
        "Ferry",
        "Field",
        "Fields",
        "Flat",
        "Flats",
        "Ford",
        "Fords",
        "Forest",
        "Forge",
        "Forges",
        "Fork",
        "Forks",
        "Fort",
        "Freeway",
        "Garden",
        "Gardens",
        "Gateway",
        "Glen",
        "Glens",
        "Green",
        "Greens",
        "Grove",
        "Groves",
        "Harbor",
        "Harbors",
        "Haven",
        "Heights",
        "Highway",
        "Hill",
        "Hills",
        "Hollow",
        "Inlet",
        "Inlet",
        "Island",
        "Island",
        "Islands",
        "Islands",
        "Isle",
        "Isle",
        "Junction",
        "Junctions",
        "Key",
        "Keys",
        "Knoll",
        "Knolls",
        "Lake",
        "Lakes",
        "Land",
        "Landing",
        "Lane",
        "Light",
        "Lights",
        "Loaf",
        "Lock",
        "Locks",
        "Locks",
        "Lodge",
        "Lodge",
        "Loop",
        "Mall",
        "Manor",
        "Manors",
        "Meadow",
        "Meadows",
        "Mews",
        "Mill",
        "Mills",
        "Mission",
        "Mission",
        "Motorway",
        "Mount",
        "Mountain",
        "Mountain",
        "Mountains",
        "Mountains",
        "Neck",
        "Orchard",
        "Oval",
        "Overpass",
        "Park",
        "Parks",
        "Parkway",
        "Parkways",
        "Pass",
        "Passage",
        "Path",
        "Pike",
        "Pine",
        "Pines",
        "Place",
        "Plain",
        "Plains",
        "Plains",
        "Plaza",
        "Plaza",
        "Point",
        "Points",
        "Port",
        "Port",
        "Ports",
        "Ports",
        "Prairie",
        "Prairie",
        "Radial",
        "Ramp",
        "Ranch",
        "Rapid",
        "Rapids",
        "Rest",
        "Ridge",
        "Ridges",
        "River",
        "Road",
        "Road",
        "Roads",
        "Roads",
        "Route",
        "Row",
        "Rue",
        "Run",
        "Shoal",
        "Shoals",
        "Shore",
        "Shores",
        "Skyway",
        "Spring",
        "Springs",
        "Springs",
        "Spur",
        "Spurs",
        "Square",
        "Square",
        "Squares",
        "Squares",
        "Station",
        "Station",
        "Stravenue",
        "Stravenue",
        "Stream",
        "Stream",
        "Street",
        "Street",
        "Streets",
        "Summit",
        "Summit",
        "Terrace",
        "Throughway",
        "Trace",
        "Track",
        "Trafficway",
        "Trail",
        "Trail",
        "Tunnel",
        "Tunnel",
        "Turnpike",
        "Turnpike",
        "Underpass",
        "Union",
        "Unions",
        "Valley",
        "Valleys",
        "Via",
        "Viaduct",
        "View",
        "Views",
        "Village",
        "Village",
        "Villages",
        "Ville",
        "Vista",
        "Vista",
        "Walk",
        "Walks",
        "Wall",
        "Way",
        "Ways",
        "Well",
        "Wells",
    )

    postal_code_formats = ("?%? %?%", "?%?%?%")

    provinces = (
        "Alberta",
        "British Columbia",
        "Manitoba",
        "New Brunswick",
        "Newfoundland and Labrador",
        "Northwest Territories",
        "Nova Scotia",
        "Nunavut",
        "Ontario",
        "Prince Edward Island",
        "Quebec",
        "Saskatchewan",
        "Yukon Territory",
    )

    provinces_abbr = (
        "AB",
        "BC",
        "MB",
        "NB",
        "NL",
        "NT",
        "NS",
        "NU",
        "ON",
        "PE",
        "QC",
        "SK",
        "YT",
    )

    provinces_postcode_prefixes = {
        "NL": ["A"],
        "NS": ["B"],
        "PE": ["C"],
        "NB": ["E"],
        "QC": ["G", "H", "J"],
        "ON": ["K", "L", "M", "N", "P"],
        "MB": ["R"],
        "SK": ["S"],
        "AB": ["T"],
        "BC": ["V"],
        "NU": ["X"],
        "NT": ["X"],
        "YT": ["Y"],
    }

    city_formats: ElementsType[str] = (
        "{{city_prefix}} {{first_name}}{{city_suffix}}",
        "{{city_prefix}} {{first_name}}",
        "{{first_name}}{{city_suffix}}",
        "{{last_name}}{{city_suffix}}",
    )
    street_name_formats = (
        "{{first_name}} {{street_suffix}}",
        "{{last_name}} {{street_suffix}}",
    )
    street_address_formats = (
        "{{building_number}} {{street_name}}",
        "{{building_number}} {{street_name}} {{secondary_address}}",
    )
    address_formats = ("{{street_address}}\n{{city}}, {{province_abbr}} {{postalcode}}",)
    secondary_address_formats = ("Apt. ###", "Suite ###")

    def administrative_unit(self) -> str:
        """ """
        return self.random_element(self.provinces)

    province = administrative_unit

    def province_abbr(self) -> str:
        return self.random_element(self.provinces_abbr)

    def city_prefix(self) -> str:
        return self.random_element(self.city_prefixes)

    def secondary_address(self) -> str:
        return self.numerify(self.random_element(self.secondary_address_formats))

    def postal_code_letter(self) -> str:
        """
        Returns a random letter from the list of allowable
        letters in a canadian postal code
        """
        return self.random_element(self.postal_code_letters)

    def _postcode_replace(self, postal_code_format: str) -> str:
        """
        Replaces all question mark ('?') occurrences with a random letter
        from given postal_code_format, then passes result to numerify to insert
        numbers
        """
        temp = re.sub(r"\?", lambda x: self.postal_code_letter(), postal_code_format)
        return self.numerify(temp)

    def postcode(self) -> str:
        """
        Returns a random postcode
        """
        return self._postcode_replace(self.random_element(self.postal_code_formats))

    def postcode_in_province(self, province_abbr: Optional[str] = None) -> str:
        """
        Returns a random postcode within the provided province abbreviation
        """
        if province_abbr is None:
            province_abbr = self.random_element(self.provinces_abbr)

        if province_abbr in self.provinces_abbr:
            postal_code_format: str = self.random_element(self.postal_code_formats)
            postal_code_format = postal_code_format.replace(
                "?",
                self.generator.random_element(self.provinces_postcode_prefixes[province_abbr]),
                1,
            )
            return self._postcode_replace(postal_code_format)
        else:
            raise Exception("Province Abbreviation not found in list")

    def postalcode_in_province(self, province_abbr: Optional[str] = None) -> str:
        return self.postcode_in_province(province_abbr)

    def postalcode(self) -> str:
        return self.postcode()



import unittest

from faker import Faker


class TestProfileProvider(unittest.TestCase):
    """Test profile provider methods"""

    num_samples = 10

    def setUp(self):
        self.fake = Faker()
        Faker.seed(0)

    def test_simple_profile(self):
        for _ in range(self.num_samples):
            profile = self.fake.simple_profile()
            assert isinstance(profile, dict)
            assert len(profile["username"]) >= 1
            assert profile["sex"] in ["F", "M"]

        profile = self.fake.simple_profile(sex="F")
        assert profile["sex"] == "F"
        profile = self.fake.simple_profile(sex="M")
        assert profile["sex"] == "M"

    def test_profile(self):
        for _ in range(self.num_samples):
            profile = self.fake.profile()
            assert isinstance(profile, dict)
            assert len(profile["username"]) >= 1
            assert profile["sex"] in ["F", "M"]
            assert "website" in profile.keys()

        profile = self.fake.profile(sex="F")
        assert profile["sex"] == "F"
        profile = self.fake.profile(sex="M")
        assert profile["sex"] == "M"

        profile = self.fake.profile(fields=["ssn", "name"])
        assert len(profile) == 2
        assert "ssn" in profile.keys()
        assert "name" in profile.keys()

        profile = self.fake.profile(fields=[])
        assert len(profile) > 0
        assert "ssn" in profile.keys()

        profile = self.fake.profile(fields=["secret_org"])
        assert len(profile) == 0



import re

from datetime import datetime
from typing import Pattern
from unittest.mock import patch

import pytest

from faker.providers.company.az_AZ import Provider as AzAzCompanyProvider
from faker.providers.company.el_GR import Provider as ElGrCompanyProvider
from faker.providers.company.en_PH import Provider as EnPhCompanyProvider
from faker.providers.company.es_ES import Provider as EsEsCompanyProvider
from faker.providers.company.fil_PH import Provider as FilPhCompanyProvider
from faker.providers.company.hu_HU import Provider as HuHuCompanyProvider
from faker.providers.company.hy_AM import Provider as HyAmCompanyProvider
from faker.providers.company.it_IT import Provider as ItItCompanyProvider
from faker.providers.company.ja_JP import Provider as JaJpCompanyProvider
from faker.providers.company.nl_BE import Provider as NlBeCompanyProvider
from faker.providers.company.nl_NL import Provider as NlNlCompanyProvider
from faker.providers.company.pl_PL import Provider as PlPlCompanyProvider
from faker.providers.company.pl_PL import company_vat_checksum, local_regon_checksum, regon_checksum
from faker.providers.company.pt_BR import company_id_checksum
from faker.providers.company.ro_RO import Provider as RoRoCompanyProvider
from faker.providers.company.ru_RU import Provider as RuRuCompanyProvider
from faker.providers.company.ru_RU import calculate_checksum
from faker.providers.company.th_TH import Provider as ThThCompanyProvider
from faker.providers.company.tr_TR import Provider as TrTrCompanyProvider
from faker.utils.checksums import luhn_checksum


class TestAzAz:
    """Test az_AZ company provider methods"""

    def test_company_suffix(self, faker, num_samples):
        for _ in range(num_samples):
            suffix = faker.company_suffix()
            assert isinstance(suffix, str)
            assert suffix in AzAzCompanyProvider.company_suffixes

    def test_large_companies(self, faker, num_samples):
        for _ in range(num_samples):
            company = faker.large_company()
            assert isinstance(company, str)
            assert company in AzAzCompanyProvider.large_companies


class TestFiFi:
    """Test fi_FI company provider methods"""

    def _has_valid_checksum(self, company_id):
        factors = [7, 9, 10, 5, 8, 4, 2]
        checksum = 0
        for x, y in zip(company_id[:-2], factors):
            checksum += int(x) * y
        checksum %= 11
        checksum = 11 - checksum if checksum else 0
        return int(company_id[-1]) == checksum

    def test_company_business_id(self, faker, num_samples):
        for _ in range(num_samples):
            company_id = faker.company_business_id()
            assert len(company_id) == 9
            assert self._has_valid_checksum(company_id)


class TestFrFr:
    """Test fr_FR company provider methods"""

    def test_siren(self, faker, num_samples):
        for _ in range(num_samples):
            siren = faker.siren()
            assert len(siren) == 11
            assert luhn_checksum(siren.replace(" ", "")) == 0

    def test_siret(self, faker, num_samples):
        for _ in range(num_samples):
            siret = faker.siret()
            assert len(siret) == 17
            assert luhn_checksum(siret.replace(" ", "")) == 0


class TestHyAm:
    """Test hy_AM company provider methods"""

    def test_bs(self, faker, num_samples):
        for _ in range(num_samples):
            bs = faker.bs()
            assert isinstance(bs, str)

    def test_catch_phrase(self, faker, num_samples):
        for _ in range(num_samples):
            catch_phrase = faker.catch_phrase()
            assert isinstance(catch_phrase, str)

    def test_company(self, faker, num_samples):
        for _ in range(num_samples):
            company = faker.company()
            assert isinstance(company, str)

    def test_company_suffix(self, faker, num_samples):
        for _ in range(num_samples):
            suffix = faker.company_suffix()
            assert isinstance(suffix, str)
            assert suffix in HyAmCompanyProvider.company_suffixes


class TestJaJp:
    """Test ja_JP company provider methods"""

    def test_company_prefix(self, faker, num_samples):
        for _ in range(num_samples):
            prefix = faker.company_prefix()
            assert isinstance(prefix, str)
            assert prefix in JaJpCompanyProvider.company_prefixes

    def test_company_category(self, faker, num_samples):
        for _ in range(num_samples):
            category = faker.company_category()
            assert isinstance(category, str)
            assert category in JaJpCompanyProvider.company_categories

    def test_company(self, faker, num_samples):
        for _ in range(num_samples):
            company = faker.company()
            assert isinstance(company, str)
            assert any(
                company.startswith(prefix) or company.endswith(prefix)
                for prefix in JaJpCompanyProvider.company_prefixes
            )
            assert any(category in company for category in JaJpCompanyProvider.company_categories)


class TestPtBr:
    """Test pt_BR company provider methods"""

    def test_company_id_checksum(self):
        assert company_id_checksum([9, 4, 9, 5, 3, 4, 4, 1, 0, 0, 0, 1]) == [5, 1]
        assert company_id_checksum([1, 6, 0, 0, 4, 6, 3, 9, 0, 0, 0, 1]) == [8, 5]

    def test_company_id(self, faker, num_samples):
        for _ in range(num_samples):
            company_id = faker.company_id()
            assert re.fullmatch(r"\d{14}", company_id)

    def test_cnpj(self, faker, num_samples):
        for _ in range(num_samples):
            cnpj = faker.cnpj()
            assert re.fullmatch(r"\d{2}\.\d{3}\.\d{3}/0001-\d{2}", cnpj)


class TestHuHu:
    """Test hu_HU company provider methods"""

    def test_company_suffix(self, faker, num_samples):
        for _ in range(num_samples):
            suffix = faker.company_suffix()
            assert isinstance(suffix, str)
            assert suffix in HuHuCompanyProvider.company_suffixes

    def test_company(self, faker, num_samples):
        for _ in range(num_samples):
            company = faker.company()
            assert isinstance(company, str)
            assert company.split(" ")[-1] in HuHuCompanyProvider.company_suffixes


class TestPlPl:
    """Test pl_PL company provider methods"""

    def test_regon_checksum(self):
        assert regon_checksum([1, 2, 3, 4, 5, 6, 7, 8]) == 5
        assert regon_checksum([8, 9, 1, 9, 5, 7, 8, 8]) == 3
        assert regon_checksum([2, 1, 7, 1, 5, 4, 8, 3]) == 8
        assert regon_checksum([7, 9, 3, 5, 4, 7, 9, 3]) == 9
        assert regon_checksum([9, 1, 5, 9, 6, 9, 4, 7]) == 7

    def test_regon(self, faker, num_samples):
        for _ in range(num_samples):
            assert re.fullmatch(r"\d{9}", faker.regon())

    def test_local_regon_checksum(self):
        assert local_regon_checksum([1, 2, 3, 4, 5, 6, 7, 8, 5, 1, 2, 3, 4]) == 7
        assert local_regon_checksum([6, 1, 1, 9, 4, 8, 8, 3, 2, 7, 5, 8, 0]) == 3
        assert local_regon_checksum([8, 9, 2, 0, 0, 3, 6, 6, 0, 7, 0, 3, 2]) == 3
        assert local_regon_checksum([3, 5, 7, 7, 1, 0, 2, 2, 2, 5, 4, 3, 3]) == 0
        assert local_regon_checksum([9, 3, 5, 3, 1, 1, 0, 1, 2, 4, 8, 8, 2]) == 1

    def test_local_regon(self, faker, num_samples):
        for _ in range(num_samples):
            assert re.fullmatch(r"\d{14}", faker.local_regon())

    def test_company_vat_checksum(self):
        assert company_vat_checksum([7, 7, 5, 7, 7, 7, 6, 0, 5]) == 9
        assert company_vat_checksum([1, 8, 6, 5, 4, 9, 9, 6, 4]) == 2
        assert company_vat_checksum([7, 1, 2, 8, 9, 2, 4, 9, 9]) == 7
        assert company_vat_checksum([3, 5, 4, 6, 1, 0, 6, 5, 8]) == 4
        assert company_vat_checksum([3, 1, 9, 5, 5, 7, 0, 4, 5]) == 0

    def test_company_vat(self, faker, num_samples):
        for _ in range(num_samples):
            assert re.fullmatch(r"\d{10}", faker.company_vat())

    def test_company_prefix(self, faker, num_samples):
        for _ in range(num_samples):
            prefix = faker.company_prefix()
            assert isinstance(prefix, str)
            assert prefix in PlPlCompanyProvider.company_prefixes

    def test_company_suffix(self, faker, num_samples):
        for _ in range(num_samples):
            suffix = faker.company_suffix()
            assert isinstance(suffix, str)
            assert suffix in PlPlCompanyProvider.company_suffixes


class TestNlNl:
    """Test nl_NL company provider methods"""

    def test_company_prefix(self, faker, num_samples):
        for _ in range(num_samples):
            prefix = faker.company_prefix()
            assert isinstance(prefix, str)
            assert prefix in NlNlCompanyProvider.company_prefixes

    def test_company_suffix(self, faker, num_samples):
        for _ in range(num_samples):
            suffix = faker.company_suffix()
            assert isinstance(suffix, str)
            assert suffix in NlNlCompanyProvider.company_suffixes

    def test_large_companies(self, faker, num_samples):
        for _ in range(num_samples):
            company = faker.large_company()
            assert isinstance(company, str)
            assert company in NlNlCompanyProvider.large_companies


class TestEnPh:
    """Test en_PH company provider methods"""

    @classmethod
    def setup_class(cls):
        cls.company_types = EnPhCompanyProvider.company_types
        cls.company_suffixes = EnPhCompanyProvider.company_suffixes.keys()
        cls.company_products = EnPhCompanyProvider.company_products
        cls.national_corporation_pattern: Pattern = re.compile(r"^National (.*?) Corporation of the Philippines$")

    def test_random_company_noun_chain(self, faker, num_samples):
        for _ in range(num_samples):
            noun_list = faker.random_company_noun_chain().split()
            assert 1 <= len(noun_list) <= 2

    def test_random_company_acronym(self, faker, num_samples):
        for _ in range(num_samples):
            acronym = faker.random_company_acronym()
            assert 2 <= len(acronym) <= 4

    def test_company(self, faker, num_samples):
        for _ in range(num_samples):
            company = faker.company()
            if company.split()[-1] in self.company_suffixes and company.split()[-2] in self.company_types:
                continue
            else:
                national_corporation_match = self.national_corporation_pattern.fullmatch(company)
                assert national_corporation_match and national_corporation_match.group(1) in self.company_products


class TestFilPh(TestEnPh):
    """Test fil_PH company provider methods"""

    def test_PH_random_good_service_adjective_chain(self, faker, num_samples):
        for _ in range(num_samples):
            adjectives = faker.random_good_service_adjective_chain().split(" at ")
            assert all(adjective in FilPhCompanyProvider.good_service_adjectives for adjective in adjectives)


class TestTlPh(TestFilPh):
    """Test tl_PH company provider methods"""

    pass


class TestRuRu:
    """Test ru_RU company provider methods"""

    def test_calculate_checksum_nine_digits(self):
        assert calculate_checksum("164027304") == "7"
        assert calculate_checksum("629082979") == "0"
        assert calculate_checksum("0203184580") == "5"
        assert calculate_checksum("1113145630") == "0"
        assert calculate_checksum("70517081385") == "1"
        assert calculate_checksum("60307390550") == "0"

    def test_businesses_inn(self, faker, num_samples):
        for _ in range(num_samples):
            inn = faker.businesses_inn()
            assert len(inn) == 10
            assert calculate_checksum(inn[:9]) == inn[9]

    def test_individuals_inn(self, faker, num_samples):
        for _ in range(num_samples):
            inn = faker.individuals_inn()
            assert len(inn) == 12
            assert calculate_checksum(inn[:10]) == inn[10]
            assert calculate_checksum(inn[:11]) == inn[11]

    def test_businesses_ogrn(self, faker, num_samples):
        max_year = datetime.now().year - 2000
        for _ in range(num_samples):
            ogrn = faker.businesses_ogrn()
            assert len(ogrn) == 13
            assert ogrn[0] in ("1", "5")
            assert 1 <= int(ogrn[1:3]) <= max_year
            assert 1 <= int(ogrn[3:5]) <= 92
            assert int(ogrn[:-1]) % 11 % 10 == int(ogrn[-1])

    def test_individuals_ogrn(self, faker, num_samples):
        max_year = datetime.now().year - 2000
        for _ in range(num_samples):
            ogrn = faker.individuals_ogrn()
            assert len(ogrn) == 15
            assert ogrn[0] == "3"
            assert 1 <= int(ogrn[1:3]) <= max_year
            assert 1 <= int(ogrn[3:5]) <= 92
            assert int(ogrn[:-1]) % 13 % 10 == int(ogrn[-1])

    def test_kpp(self, faker, num_samples):
        for _ in range(num_samples):
            kpp = faker.kpp()
            assert len(kpp) == 9
            assert 1 <= int(kpp[0:2]) <= 92
            assert int(kpp[2:4]) > 0
            assert kpp[4:6] in ("01", "43", "44", "45")

    def test_company_prefix(self, faker, num_samples):
        for _ in range(num_samples):
            prefix = faker.company_prefix()
            assert isinstance(prefix, str)
            assert prefix in RuRuCompanyProvider.company_prefixes

    def test_company_suffix(self, faker, num_samples):
        for _ in range(num_samples):
            suffix = faker.company_suffix()
            assert isinstance(suffix, str)
            assert suffix in RuRuCompanyProvider.company_suffixes

    def test_large_companies(self, faker, num_samples):
        for _ in range(num_samples):
            company = faker.large_company()
            assert isinstance(company, str)
            assert company in RuRuCompanyProvider.large_companies

    def test_catchphrase(self, faker, num_samples):
        for _ in range(num_samples):
            catchphrase = faker.catch_phrase()
            assert isinstance(catchphrase, str)
            assert " и " in catchphrase

    def test_bs(self, faker, num_samples):
        for _ in range(num_samples):
            bs = faker.bs()
            bs_words = bs.split()
            assert isinstance(bs, str)
            assert bs_words[0] in RuRuCompanyProvider.bsWords[0]


class TestItIt:
    """Test it_IT company provider methods"""

    vat_regex = re.compile(r"^IT\d{7}(0\d{2}|100|120|121|888|999)\d$", flags=re.ASCII)

    def test_company_vat(self, faker, num_samples):
        for _ in range(num_samples):
            company_vat = faker.company_vat()
            assert self.vat_regex.match(company_vat)

    @pytest.mark.parametrize(
        "value, expected",
        (
            (100, "100"),
            (101, "120"),
            (102, "121"),
            (103, "888"),
            (104, "999"),
        ),
    )
    def test_company_vat_special_cases(self, faker, value, expected):
        # this test allows to get full code coverage for company_vat fixing the internal state of the random generator
        fake = ItItCompanyProvider(generator=faker)

        with patch.object(fake, "random_int", return_value=value, autospec=True):
            company_vat = fake.company_vat()
            assert self.vat_regex.match(company_vat)
            assert company_vat[9:12] == expected


class TestThTh:
    """Test th_TH company provider methods"""

    def test_company_prefix(self, faker, num_samples):
        for _ in range(num_samples):
            prefix = faker.company_prefix()
            assert isinstance(prefix, str)
            assert prefix in ThThCompanyProvider.company_prefixes

    def test_company_suffix(self, faker, num_samples):
        for _ in range(num_samples):
            suffix = faker.company_suffix()
            assert isinstance(suffix, str)
            assert suffix in ThThCompanyProvider.company_suffixes

    def test_company_limited_prefix(self, faker, num_samples):
        for _ in range(num_samples):
            prefix = faker.company_limited_prefix()
            assert isinstance(prefix, str)
            assert prefix in ThThCompanyProvider.company_limited_prefixes

    def test_company_limited_suffix(self, faker, num_samples):
        for _ in range(num_samples):
            suffix = faker.company_limited_suffix()
            assert isinstance(suffix, str)
            assert suffix in ThThCompanyProvider.company_limited_suffixes

    def test_nonprofit_prefix(self, faker, num_samples):
        for _ in range(num_samples):
            prefix = faker.nonprofit_prefix()
            assert isinstance(prefix, str)
            assert prefix in ThThCompanyProvider.nonprofit_prefixes

    def test_company(self, faker, num_samples):
        for _ in range(num_samples):
            company = faker.company()
            assert isinstance(company, str)


class TestTrTr:
    """Test tr_TR company provider methods"""

    def test_company_suffix(self, faker, num_samples):
        for _ in range(num_samples):
            suffix = faker.company_suffix()
            assert isinstance(suffix, str)
            assert suffix in TrTrCompanyProvider.company_suffixes

    def test_large_companies(self, faker, num_samples):
        for _ in range(num_samples):
            company = faker.large_company()
            assert isinstance(company, str)
            assert company in TrTrCompanyProvider.large_companies


class TestRoRo:
    """Test ro_RO company provider methods"""

    def test_company_suffix(self, faker, num_samples):
        for _ in range(num_samples):
            suffix = faker.company_suffix()
            assert isinstance(suffix, str)
            assert suffix in RoRoCompanyProvider.company_suffixes


class TestElGr:
    """Test el_GR company provider methods"""

    def test_company_suffix(self, faker, num_samples):
        for _ in range(num_samples):
            suffix = faker.company_suffix()
            assert isinstance(suffix, str)
            assert suffix in ElGrCompanyProvider.company_suffixes


class TestNlBe:
    """Test nl_BE company provider methods"""

    def test_company_suffix(self, faker, num_samples):
        for _ in range(num_samples):
            suffix = faker.company_suffix()
            assert isinstance(suffix, str)
            assert suffix in NlBeCompanyProvider.company_suffixes


class TestEsEs:
    """Test esE_ES company provider methods"""

    def test_company_suffix(self, faker, num_samples):
        for _ in range(num_samples):
            suffix = faker.company_suffix()
            assert isinstance(suffix, str)
            assert suffix in list(EsEsCompanyProvider.company_suffixes.keys())

    def test_company(self, faker, num_samples):
        for _ in range(num_samples):
            company = faker.company()
            assert isinstance(company, str)



from .. import Provider as PhoneNumberProvider


class Provider(PhoneNumberProvider):
    """
    According to official specs:
    https://www.bcn.cl/leychile/navegar?i=173931
    https://www.itu.int/itudoc/itu-t/number/c/chl/76483_ww9-es.doc
    https://www.subtel.gob.cl/base_numeracion/tabla_numeracion_ido_idd.xlsx
    """

    formats = (
        "+56 2 2%## ####",  # santiago landline
        "+56 2 3%## ####",
        "+56 {{ landline_code }} 2%# ####",  # national landlines and VoIP
        "+56 {{ landline_code }} 3%# ####",
        "+56 9 {{ cellphone_block }}%## ####",  # cell phones
        "+56 {{ special_code }} %## ###",  # special
    )

    landline_codes = (
        "32",  # Valparaíso
        "33",  # Quillota
        "34",  # Los Andes
        "35",  # San Antonio
        "41",  # Concepción
        "42",  # Chillán
        "43",  # Los Ángeles
        "45",  # Temuco
        "51",  # La Serena
        "52",  # Copiapó
        "53",  # Ovalle
        "55",  # Antofagasta
        "57",  # Iquique
        "58",  # Arica
        "61",  # Punta Arenas
        "63",  # Valdivia
        "64",  # Osorno
        "65",  # Puerto Montt
        "67",  # Coyhaique
        "71",  # Talca
        "72",  # Rangagua
        "73",  # Linares
        "75",  # Curicó
        "44",  # VoIP
    )

    special_codes = (
        "600",  # Nationalwide
        "800",  # Nationalwide, toll-free
    )

    cellphone_blocks = ("2", "3", "4", "5", "6", "7", "8", "9")

    def landline_code(self) -> str:
        return self.numerify(self.random_element(self.landline_codes))

    def cellphone_block(self) -> str:
        return self.random_element(self.cellphone_blocks)

    def special_code(self) -> str:
        return self.numerify(self.random_element(self.special_codes))

    def phone_number(self) -> str:
        return self.numerify(self.generator.parse(self.random_element(self.formats)))



from .. import Provider as BankProvider


class Provider(BankProvider):
    """Implement bank provider for ``zh_CN`` locale.
    Source: https://zh.wikipedia.org/wiki/中国大陆银行列表
    """

    banks = (
        "中国人民银行",
        "国家开发银行",
        "中国进出口银行",
        "中国农业发展银行",
        "交通银行",
        "中国银行",
        "中国建设银行",
        "中国农业银行",
        "中国工商银行",
        "中国邮政储蓄银行",
        "中国光大银行",
        "中国民生银行",
        "招商银行",
        "中信银行",
        "华夏银行",
        "上海浦东发展银行",
        "平安银行",
        "广发银行",
        "兴业银行",
        "浙商银行",
        "渤海银行",
        "恒丰银行",
        "西安银行",
    )

    def bank(self) -> str:
        """Generate a bank name."""
        return self.random_element(self.banks)



from ..ar_AA import Provider as ArabicPersonProvider


class Provider(ArabicPersonProvider):
    last_names = (
        "آل الشيخ",
        "آل العسكري",
        "آل بن ظافر",
        "آل بن لافي",
        "آل جعفر",
        "آل حسين",
        "آل خضير",
        "آل رفيع",
        "آل سعود",
        "آل سلطان",
        "آل صفوان",
        "آل عايض",
        "آل عطفة",
        "آل علي",
        "آل عواض",
        "آل قصير",
        "آل محمد بن علي بن جماز",
        "آل معيض",
        "آل مقطة",
        "أبا الخيل",
        "أبو داوود",
        "الجابر",
        "الجفالي",
        "الحجار",
        "الحكير",
        "الخرافي",
        "الدباغ",
        "الراشد",
        "الشايع",
        "العجلان",
        "العقيل",
        "العليان",
        "المشاولة",
        "المغاولة",
        "المهنا",
        "المهيدب",
        "بقشان",
        "بن لادن",
        "حجار",
        "حنبولي",
        "شربتلي",
        "فصيل",
        "كانو",
        "مهنا",
    )



import locale as pylocale
import logging
import sys

from importlib import import_module
from typing import Any, List, Optional, Tuple

from .config import AVAILABLE_LOCALES, DEFAULT_LOCALE, PROVIDERS
from .generator import Generator
from .utils.loading import list_module

logger = logging.getLogger(__name__)

# identify if python is being run in interactive mode. If so, disable logging.
inREPL = bool(getattr(sys, "ps1", False))
if inREPL:
    logger.setLevel(logging.CRITICAL)
else:
    logger.debug("Not in REPL -> leaving logger event level as is.")


class Factory:
    @classmethod
    def create(
        cls,
        locale: Optional[str] = None,
        providers: Optional[List[str]] = None,
        generator: Optional[Generator] = None,
        includes: Optional[List[str]] = None,
        # Should we use weightings (more realistic) or weight every element equally (faster)?
        # By default, use weightings for backwards compatibility & realism
        use_weighting: bool = True,
        **config: Any,
    ) -> Generator:
        if includes is None:
            includes = []

        # fix locale to package name
        locale = locale.replace("-", "_") if locale else DEFAULT_LOCALE
        locale = pylocale.normalize(locale).split(".")[0]
        if locale not in AVAILABLE_LOCALES:
            msg = f"Invalid configuration for faker locale `{locale}`"
            raise AttributeError(msg)

        config["locale"] = locale
        config["use_weighting"] = use_weighting
        providers = providers or PROVIDERS

        providers += includes

        faker = generator or Generator(**config)

        for prov_name in providers:
            if prov_name == "faker.providers":
                continue

            prov_cls, lang_found, _ = cls._find_provider_class(prov_name, locale)
            provider = prov_cls(faker)
            provider.__use_weighting__ = use_weighting
            provider.__provider__ = prov_name
            provider.__lang__ = lang_found
            faker.add_provider(provider)

        return faker

    @classmethod
    def _find_provider_class(
        cls,
        provider_path: str,
        locale: Optional[str] = None,
    ) -> Tuple[Any, Optional[str], Optional[str]]:
        provider_module = import_module(provider_path)
        default_locale = getattr(provider_module, "default_locale", "")

        if getattr(provider_module, "localized", False):
            logger.debug(
                "Looking for locale `%s` in provider `%s`.",
                locale,
                provider_module.__name__,
            )

            available_locales = list_module(provider_module)
            if not locale or locale not in available_locales:
                unavailable_locale = locale
                locale = default_locale or DEFAULT_LOCALE
                logger.debug(
                    "Specified locale `%s` is not available for "
                    "provider `%s`. Locale reset to `%s` for this "
                    "provider.",
                    unavailable_locale,
                    provider_module.__name__,
                    locale,
                )
            else:
                logger.debug(
                    "Provider `%s` has been localized to `%s`.",
                    provider_module.__name__,
                    locale,
                )

            path = f"{provider_path}.{locale}"
            provider_module = import_module(path)

        else:
            if locale:
                logger.debug(
                    "Provider `%s` does not feature localization. "
                    "Specified locale `%s` is not utilized for this "
                    "provider.",
                    provider_module.__name__,
                    locale,
                )
            locale = default_locale = None

        return provider_module.Provider, locale, default_locale  # type: ignore



from .. import Provider as PersonProvider


class Provider(PersonProvider):
    formats = ("{{first_name}} {{last_name}}",)

    # First names are from
    #   https://www.babycenter.in/a25010193/modern-indian-baby-names
    #   https://en.wikipedia.org/wiki/Category:Male_actors_in_Malayalam_cinema (not used exhaustively)
    #   https://en.wikipedia.org/wiki/List_of_Tamil_film_actors (not used exhaustively)
    # Last names are from https://www.familyeducation.com/baby-names/browse-origin/surname/indian

    first_names = (
        "Aarav",
        "Aarush",
        "Aayush",
        "Abram",
        "Advik",
        "Akarsh",
        "Anay",
        "Aniruddh",
        "Arhaan",
        "Armaan",
        "Arnav",
        "Azad",
        "Badal",
        "Baiju",
        "Bhavin",
        "Biju",
        "Chirag",
        "Darshit",
        "Devansh",
        "Dhanuk",
        "Dhanush",
        "Dharmajan",
        "Dhruv",
        "Divij",
        "Divit",
        "Divyansh",
        "Ehsaan",
        "Emir",
        "Faiyaz",
        "Farhan",
        "Fateh",
        "Gatik",
        "Gokul",
        "Hansh",
        "Himmat",
        "Hiran",
        "Hridaan",
        "Hunar",
        "Indrajit",
        "Indrans",
        "Indranil",
        "Ishaan",
        "Ivan",
        "Jayan",
        "Jayant",
        "Jayesh",
        "Jivin",
        "Kabir",
        "Kanav",
        "Kartik",
        "Kiaan",
        "Krish",
        "Lagan",
        "Lakshay",
        "Lakshit",
        "Madhav",
        "Madhup",
        "Mamooty",
        "Manikya",
        "Mehul",
        "Miraan",
        "Nakul",
        "Nirvaan",
        "Nishith",
        "Ojas",
        "Onkar",
        "Pranay",
        "Prerak",
        "Priyansh",
        "Purab",
        "Raghav",
        "Ranbir",
        "Raunak",
        "Reyansh",
        "Riaan",
        "Ritvik",
        "Rohan",
        "Romil",
        "Ryan",
        "Sahil",
        "Saksham",
        "Samar",
        "Samarth",
        "Shaan",
        "Shalv",
        "Shamik",
        "Shayak",
        "Shlok",
        "Shray",
        "Stuvan",
        "Sumer",
        "Taimur",
        "Taran",
        "Tejas",
        "Tushar",
        "Umang",
        "Uthkarsh",
        "Vaibhav",
        "Veer",
        "Vidur",
        "Vihaan",
        "Vivaan",
        "Yakshit",
        "Yuvaan",
        "Yuvraj ",
        "Zain",
        "Zeeshan",
        "Aaina",
        "Aarna",
        "Aaryahi",
        "Adah",
        "Adira",
        "Advika",
        "Ahana ",
        "Alia",
        "Alisha",
        "Amani",
        "Amira",
        "Anahi",
        "Anahita",
        "Anaya",
        "Anika",
        "Anvi",
        "Anya",
        "Aradhya",
        "Ayesha",
        "Bhamini",
        "Charvi",
        "Damini",
        "Dishani",
        "Diya",
        "Drishya",
        "Ela",
        "Elakshi",
        "Eshani",
        "Eva",
        "Hazel",
        "Heer",
        "Hrishita",
        "Inaaya ",
        "Ira",
        "Ishita",
        "Ivana",
        "Jhanvi",
        "Jivika",
        "Jiya",
        "Kaira",
        "Kashvi",
        "Kavya",
        "Keya",
        "Khushi",
        "Kiara",
        "Kimaya",
        "Kismat",
        "Lavanya",
        "Mahika",
        "Manjari",
        "Mannat",
        "Miraya",
        "Misha",
        "Mishti",
        "Mohanlal",
        "Myra",
        "Navya",
        "Nayantara",
        "Neelofar",
        "Nehmat",
        "Neysa",
        "Nirvi",
        "Nitara",
        "Nitya",
        "Oorja",
        "Pari",
        "Parinaaz",
        "Pihu",
        "Piya",
        "Prisha",
        "Rania",
        "Rasha",
        "Rati",
        "Renee",
        "Rhea",
        "Riya",
        "Saanvi",
        "Saira",
        "Samaira",
        "Samiha",
        "Sana",
        "Sara",
        "Seher",
        "Shanaya",
        "Siya",
        "Suhana",
        "Tanya",
        "Tara",
        "Tarini",
        "Tiya",
        "Trisha",
        "Urvi",
        "Vanya",
        "Vardaniya",
        "Vedika",
        "Vritika",
        "Yashvi",
        "Yasmin",
        "Zaina",
        "Zara",
        "Zoya",
    )

    last_names = (
        "Acharya",
        "Agarwal",
        "Agate",
        "Aggarwal",
        "Agrawal",
        "Ahluwalia",
        "Ahuja",
        "Amble",
        "Anand",
        "Andra",
        "Anne",
        "Apte",
        "Arora",
        "Arya",
        "Atwal",
        "Aurora",
        "Babu",
        "Badal",
        "Badami",
        "Bahl",
        "Bahri",
        "Bail",
        "Bains",
        "Bajaj",
        "Bajwa",
        "Bakshi",
        "Bal",
        "Bala",
        "Bala",
        "Balakrishnan",
        "Balan",
        "Balasubramanian",
        "Balay",
        "Bali",
        "Bandi",
        "Banerjee",
        "Banik",
        "Bansal",
        "Barad",
        "Barad",
        "Baral",
        "Baria",
        "Barman",
        "Basak",
        "Bassi",
        "Basu",
        "Bath",
        "Batra",
        "Batta",
        "Bava",
        "Bawa",
        "Bedi",
        "Behl",
        "Ben",
        "Bera",
        "Bhagat",
        "Bhakta",
        "Bhalla",
        "Bhandari",
        "Bhardwaj",
        "Bhargava",
        "Bhasin",
        "Bhat",
        "Bhatia",
        "Bhatnagar",
        "Bhatt",
        "Bhattacharyya",
        "Bhatti",
        "Bhavsar",
        "Bir",
        "Biswas",
        "Boase",
        "Bobal",
        "Bora",
        "Bora",
        "Borah",
        "Borde",
        "Borra",
        "Bose",
        "Brahmbhatt",
        "Brar",
        "Buch",
        "Buch",
        "Bumb",
        "Butala",
        "Chacko",
        "Chad",
        "Chada",
        "Chadha",
        "Chahal",
        "Chakrabarti",
        "Chakraborty",
        "Chana",
        "Chand",
        "Chanda",
        "Chander",
        "Chandra",
        "Chandran",
        "Char",
        "Chatterjee",
        "Chaudhari",
        "Chaudhary",
        "Chaudhry",
        "Chaudhuri",
        "Chaudry",
        "Chauhan",
        "Chawla",
        "Cheema",
        "Cherian",
        "Chhabra",
        "Chokshi",
        "Chopra",
        "Choudhary",
        "Choudhry",
        "Choudhury",
        "Chowdhury",
        "Comar",
        "Contractor",
        "D’Alia",
        "Dada",
        "Dalal",
        "Dani",
        "Dar",
        "Dara",
        "Dara",
        "Das",
        "Dasgupta",
        "Dash",
        "Dass",
        "Date",
        "Datta",
        "Dave",
        "Dayal",
        "De",
        "Deep",
        "Deo",
        "Deol",
        "Desai",
        "Deshmukh",
        "Deshpande",
        "Devan",
        "Devi",
        "Dewan",
        "Dey",
        "Dhaliwal",
        "Dhar",
        "Dhar",
        "Dhawan",
        "Dhillon",
        "Dhingra",
        "Din",
        "Divan",
        "Dixit",
        "Doctor",
        "Dora",
        "Doshi",
        "Dua",
        "Dube",
        "Dubey",
        "Dugal",
        "Dugar",
        "Dugar",
        "Dutt",
        "Dutta",
        "Dyal",
        "Edwin",
        "Gaba",
        "Gade",
        "Gala",
        "Gandhi",
        "Ganesan",
        "Ganesh",
        "Ganguly",
        "Gara",
        "Garde",
        "Garg",
        "Gera",
        "Ghose",
        "Ghosh",
        "Gill",
        "Goda",
        "Goel",
        "Gokhale",
        "Gola",
        "Gole",
        "Golla",
        "Gopal",
        "Goswami",
        "Gour",
        "Goyal",
        "Grewal",
        "Grover",
        "Guha",
        "Gulati",
        "Gupta",
        "Halder",
        "Handa",
        "Hans",
        "Hari",
        "Hayer",
        "Hayre",
        "Hegde",
        "Hora",
        "Issac",
        "Iyengar",
        "Iyer",
        "Jaggi",
        "Jain",
        "Jani",
        "Jayaraman",
        "Jha",
        "Jhaveri",
        "Johal",
        "Joshi",
        "Kadakia",
        "Kade",
        "Kakar",
        "Kala",
        "Kala",
        "Kala",
        "Kale",
        "Kalita",
        "Kalla",
        "Kamdar",
        "Kanda",
        "Kannan",
        "Kant",
        "Kapadia",
        "Kapoor",
        "Kapur",
        "Kar",
        "Kara",
        "Karan",
        "Kari",
        "Karnik",
        "Karpe",
        "Kashyap",
        "Kata",
        "Kaul",
        "Kaur",
        "Keer",
        "Keer",
        "Khalsa",
        "Khanna",
        "Khare",
        "Khatri",
        "Khosla",
        "Khurana",
        "Kibe",
        "Kohli",
        "Konda",
        "Korpal",
        "Koshy",
        "Kota",
        "Kothari",
        "Krish",
        "Krishna",
        "Krishnamurthy",
        "Krishnan",
        "Kulkarni",
        "Kumar",
        "Kumer",
        "Kunda",
        "Kurian",
        "Kuruvilla",
        "Lad",
        "Lad",
        "Lal",
        "Lala",
        "Lall",
        "Lalla",
        "Lanka",
        "Lata",
        "Loke",
        "Loyal",
        "Luthra",
        "Madan",
        "Madan",
        "Magar",
        "Mahajan",
        "Mahal",
        "Maharaj",
        "Majumdar",
        "Malhotra",
        "Mall",
        "Mallick",
        "Mammen",
        "Mand",
        "Manda",
        "Mandal",
        "Mander",
        "Mane",
        "Mangal",
        "Mangat",
        "Mani",
        "Mani",
        "Mann",
        "Mannan",
        "Manne",
        "Master",
        "Raj",
        "Raja",
        "Rajagopal",
        "Rajagopalan",
        "Rajan",
        "Raju",
        "Ram",
        "Rama",
        "Ramachandran",
        "Ramakrishnan",
        "Raman",
        "Ramanathan",
        "Ramaswamy",
        "Ramesh",
        "Rana",
        "Randhawa",
        "Ranganathan",
        "Rao",
        "Rastogi",
        "Ratta",
        "Rattan",
        "Ratti",
        "Rau",
        "Raval",
        "Ravel",
        "Ravi",
        "Ray",
        "Reddy",
        "Rege",
        "Rout",
        "Roy",
        "Sabharwal",
        "Sachar",
        "Sachdev",
        "Sachdeva",
        "Sagar",
        "Saha",
        "Sahni",
        "Sahota",
        "Saini",
        "Salvi",
        "Sama",
        "Sami",
        "Sampath",
        "Samra",
        "Sandal",
        "Sandhu",
        "Sane",
        "Sangha",
        "Sanghvi",
        "Sani",
        "Sankar",
        "Sankaran",
        "Sant",
        "Saraf",
        "Saran",
        "Sarin",
        "Sarkar",
        "Sarma",
        "Sarna",
        "Sarraf",
        "Sastry",
        "Sathe",
        "Savant",
        "Sawhney",
        "Saxena",
        "Sehgal",
        "Sekhon",
        "Sem",
        "Sen",
        "Sengupta",
        "Seshadri",
        "Seth",
        "Sethi",
        "Setty",
        "Sha",
        "Shah",
        "Shan",
        "Shankar",
        "Shanker",
        "Sharaf",
        "Sharma",
        "Shenoy",
        "Shere",
        "Sheth",
        "Shetty",
        "Shroff",
        "Shukla",
        "Sibal",
        "Sidhu",
        "Singh",
        "Singhal",
        "Sinha",
        "Sodhi",
        "Solanki",
        "Som",
        "Soman",
        "Soni",
        "Sood",
        "Sridhar",
        "Srinivas",
        "Srinivasan",
        "Srivastava",
        "Subramaniam",
        "Subramanian",
        "Sule",
        "Sundaram",
        "Sunder",
        "Sur",
        "Sura",
        "Suresh",
        "Suri",
        "Swaminathan",
        "Swamy",
        "Tailor",
        "Tak",
        "Talwar",
        "Tandon",
        "Taneja",
        "Tank",
        "Tara",
        "Tata",
        "Tella",
        "Thaker",
        "Thakkar",
        "Thakur",
        "Thaman",
        "Tiwari",
        "Toor",
        "Tripathi",
        "Trivedi",
        "Upadhyay",
        "Uppal",
        "Vaidya",
        "Vala",
        "Varghese",
        "Varkey",
        "Varma",
        "Varty",
        "Varughese",
        "Vasa",
        "Venkataraman",
        "Venkatesh",
        "Verma",
        "Vig",
        "Virk",
        "Viswanathan",
        "Vohra",
        "Vora",
        "Vyas",
        "Wable",
        "Wadhwa",
        "Wagle",
        "Wali",
        "Wali",
        "Walia",
        "Walla",
        "Warrior",
        "Wason",
        "Yadav",
        "Yogi",
        "Yohannan",
        "Zacharia",
        "Zachariah",
    )



import pytest

from faker import Faker
from faker.config import DEFAULT_LOCALE

DEFAULT_SEED = 0


@pytest.fixture(scope="session", autouse=True)
def _session_faker(request):
    """Fixture that stores the session level ``Faker`` instance.

    This fixture is internal and is only meant for use within the project.
    Third parties should instead use the ``faker`` fixture for their tests.
    """
    if "faker_session_locale" in request.fixturenames:
        locale = request.getfixturevalue("faker_session_locale")
    else:
        locale = [DEFAULT_LOCALE]
    return Faker(locale=locale)


@pytest.fixture()
def faker(request):
    """Fixture that returns a seeded and suitable ``Faker`` instance."""
    if "faker_locale" in request.fixturenames:
        locale = request.getfixturevalue("faker_locale")
        fake = Faker(locale=locale)
    else:
        fake = request.getfixturevalue("_session_faker")

    seed = DEFAULT_SEED
    if "faker_seed" in request.fixturenames:
        seed = request.getfixturevalue("faker_seed")
    fake.seed_instance(seed=seed)
    fake.unique.clear()

    return fake



from .. import Provider as BankProvider


class Provider(BankProvider):
    """Implement bank provider for ``en_IN`` locale.
    Source: https://en.wikipedia.org/wiki/List_of_banks_in_India
    """

    banks = (
        "Bank of Baroda",
        "Bank of India",
        "Bank of Maharashtra",
        "Canara Bank",
        "Central Bank of India",
        "Indian Bank",
        "Indian Overseas Bank",
        "Punjab National Bank",
        "Punjab and Sind Bank",
        "Union Bank of India",
        "UCO Bank",
        "State Bank of India",
        "Axis Bank",
        "Bandhan Bank",
        "CSB Bank",
        "City Union Bank",
        "DCB Bank",
        "Dhanlaxmi Bank",
        "Federal Bank",
        "HDFC Bank",
        "ICICI Bank",
        "IDBI Bank",
        "IDFC First Bank",
        "IndusInd Bank",
        "Jammu & Kashmir Bank",
        "Karnataka Bank",
        "Karur Vysya Bank",
        "Kotak Mahindra Bank",
        "Nainital Bank",
        "RBL Bank",
        "South Indian Bank",
        "Tamilnad Mercantile Bank",
        "Yes Bank",
    )

    def bank(self) -> str:
        """Generate a bank name."""
        return self.random_element(self.banks)



from .. import Provider as DateTimeProvider


class Provider(DateTimeProvider):
    MONTH_NAMES = {
        "01": "一月",
        "02": "二月",
        "03": "三月",
        "04": "四月",
        "05": "五月",
        "06": "六月",
        "07": "七月",
        "08": "八月",
        "09": "九月",
        "10": "十月",
        "11": "十一月",
        "12": "十二月",
    }
    DAY_NAMES = {
        "0": "星期日",
        "1": "星期一",
        "2": "星期二",
        "3": "星期三",
        "4": "星期四",
        "5": "星期五",
        "6": "星期六",
    }

    def day_of_week(self) -> str:
        day = self.date("%w")
        return self.DAY_NAMES[day]

    def month_name(self) -> str:
        month = self.month()
        return self.MONTH_NAMES[month]



"""TEST NOTES:

The following tests cover cases where a ``faker_session_locale`` fixture was not
defined by the user, but an autouse ``faker_seed`` fixture was defined. In this
setup, the plugin's ``DEFAULT_SEED`` will be ignored, and ``Faker`` instances will
be seeded using the value of ``faker_seed``. Said instances are still chosen in
accordance to how ``faker_locale`` and ``faker_session_locale`` interact with
each other.
"""

from random import Random

import pytest

from faker.contrib.pytest.plugin import DEFAULT_LOCALE

_CHANGED_SEED = 4761


@pytest.fixture()
def faker_locale():
    return ["it_IT"]


@pytest.fixture(autouse=True)
def faker_seed():
    return _CHANGED_SEED


def test_no_injection(_session_faker, faker):
    random = Random(_CHANGED_SEED)
    assert faker == _session_faker
    assert faker.locales == [DEFAULT_LOCALE]
    assert faker.random != random
    assert faker.random.getstate() == random.getstate()


def test_inject_faker_locale(_session_faker, faker, faker_locale):
    random = Random(_CHANGED_SEED)
    assert faker != _session_faker
    assert faker.locales == faker_locale
    assert faker.random != random
    assert faker.random.getstate() == random.getstate()



from .. import Provider as DateTimeProvider


class Provider(DateTimeProvider):
    DAY_NAMES = {
        "0": "poniedziałek",
        "1": "wtorek",
        "2": "środa",
        "3": "czwartek",
        "4": "piątek",
        "5": "sobota",
        "6": "niedziela",
    }

    MONTH_NAMES = {
        "01": "styczeń",
        "02": "luty",
        "03": "marzec",
        "04": "kwiecień",
        "05": "maj",
        "06": "czerwiec",
        "07": "lipiec",
        "08": "sierpień",
        "09": "wrzesień",
        "10": "październik",
        "11": "listopad",
        "12": "grudzień",
    }

    def day_of_week(self):
        day = self.date("%w")
        return self.DAY_NAMES[day]

    def month_name(self):
        month = self.month()
        return self.MONTH_NAMES[month]



import locale as pylocale
import re

import pytest

from faker import Faker
from faker.contrib.pytest.plugin import DEFAULT_SEED

LOCALE_TEST_CLASS_NAME_REGEX = re.compile(
    r"^Test(?P<language>[A-Z][a-z]{1,2})(?P<region>[A-Z][a-z])$",
)


@pytest.fixture(scope="class", autouse=True)
def _class_locale_faker(request):
    if not request.cls:
        return None
    class_name = request.cls.__name__
    match = LOCALE_TEST_CLASS_NAME_REGEX.fullmatch(class_name)
    if not match:
        return None
    locale = f'{match.group("language").lower()}_{match.group("region").upper()}'
    locale = pylocale.normalize(locale).split(".")[0]
    return Faker(locale=locale)


@pytest.fixture(autouse=True)
def faker(_class_locale_faker, faker):
    if not _class_locale_faker:
        return faker
    _class_locale_faker.seed_instance(DEFAULT_SEED)
    return _class_locale_faker


@pytest.fixture(scope="class", autouse=True)
def num_samples(request):
    try:
        num = int(request.cls.num_samples)
    except AttributeError:
        num = 100
    return num



from .. import Provider as PersonProvider


class Provider(PersonProvider):
    formats_female = (
        "{{first_name_female}} {{last_name}}",
        "{{first_name_female}} {{last_name}}",
        "{{first_name_female}} {{last_name}}",
        "{{first_name_female}} {{last_name}}",
        "{{first_name_female}} {{last_name}}",
        "{{first_name_female}} {{last_name}}",
        "{{first_name_female}} {{last_name}}-{{last_name}}",
        "{{first_name_female}}-{{first_name_female}} {{last_name}}",
    )

    formats_male = (
        "{{first_name_male}} {{last_name}}",
        "{{first_name_male}} {{last_name}}",
        "{{first_name_male}} {{last_name}}",
        "{{first_name_male}} {{last_name}}",
        "{{first_name_male}} {{last_name}}",
        "{{first_name_male}} {{last_name}}",
        "{{first_name_male}} {{last_name}}-{{last_name}}",
        "{{first_name_male}}-{{first_name_male}} {{last_name}}",
    )

    formats = formats_male + formats_female

    # Source:
    # https://www.retraitequebec.gouv.qc.ca/fr/services-en-ligne-outils/banque-de-prenoms/Pages/banque-de-prenoms.aspx
    first_names_male = (
        "Alain",
        "Alexandre",
        "Alexis",
        "André",
        "Antoine",
        "Arthur",
        "Benjamin",
        "Benoît",
        "Bernard",
        "Bertrand",
        "Charles",
        "Daniel",
        "David",
        "Denis",
        "Édouard",
        "Émile",
        "Emmanuel",
        "Éric",
        "Étienne",
        "François",
        "Frédéric",
        "Gabriel",
        "Georges",
        "Gérard",
        "Gilbert",
        "Gilles",
        "Grégoire",
        "Guillaume",
        "Guy",
        "William",
        "Henri",
        "Hugues",
        "Jacques",
        "Jean",
        "Jérôme",
        "Jonathan",
        "Joseph",
        "Jules",
        "Julien",
        "Kevin",
        "Laurent",
        "Louis",
        "Luc",
        "Lucas",
        "Marc",
        "Maxime",
        "Marcel",
        "Martin",
        "Mathieu",
        "Maurice",
        "Michel",
        "Nathan",
        "Nicolas",
        "Noël",
        "Olivier",
        "Patrick",
        "Paul",
        "Philippe",
        "Pierre",
        "Raphaël",
        "Raymond",
        "Rémy",
        "René",
        "Richard",
        "Robert",
        "Roger",
        "Roland",
        "Samuel",
        "Sébastien",
        "Stéphane",
        "Théodore",
        "Thomas",
        "Timothée",
        "Tristan",
        "Victor",
        "Vincent",
        "Xavier",
        "Yves",
        "Zacharie",
    )

    first_names_female = (
        "Agnès",
        "Alexandra",
        "Alex",
        "Alice",
        "Amélie",
        "Anaïs",
        "Andrée",
        "Anne",
        "Anouk",
        "Astrid",
        "Audrey",
        "Aurélie",
        "Aurore",
        "Béatrice",
        "Brigitte",
        "Camille",
        "Caroline",
        "Catherine",
        "Cécile",
        "Céline",
        "Célina",
        "Chantal",
        "Charlotte",
        "Christelle",
        "Christiane",
        "Christine",
        "Claire",
        "Claudine",
        "Claude",
        "Clémence",
        "Colette",
        "Constance",
        "Danielle",
        "Denise",
        "Diane",
        "Dominique",
        "Dorothée",
        "Édith",
        "Emma",
        "Éléonore",
        "Élisabeth",
        "Élise",
        "Élodie",
        "Émilie",
        "Emmanuelle",
        "Florence",
        "Françoise",
        "Frédérique",
        "Gabrielle",
        "Geneviève",
        "Hélène",
        "Henriette",
        "Hortense",
        "Isabelle",
        "Jacqueline",
        "Jeanne",
        "Jeannine",
        "Jessica",
        "Joséphine",
        "Josette",
        "Julie",
        "Juliette",
        "Karine",
        "Laetitia",
        "Laurence",
        "Laurie",
        "Lorraine",
        "Louise",
        "Lucie",
        "Lucy",
        "Manon",
        "Marcelle",
        "Marguerite",
        "Margot",
        "Margaret",
        "Marianne",
        "Marie",
        "Martine",
        "Maryse",
        "Maude",
        "Mathilde",
        "Mélanie",
        "Michèle",
        "Michelle",
        "Monique",
        "Nathalie",
        "Nathalie",
        "Nicole",
        "Noémie",
        "Océane",
        "Odette",
        "Olivia",
        "Patricia",
        "Paulette",
        "Pauline",
        "Pénélope",
        "Renée",
        "Rosalie",
        "Sarah",
        "Simone",
        "Sophie",
        "Stéphanie",
        "Susanne",
        "Sylvie",
        "Thérèse",
        "Vanessa",
        "Valérie",
        "Véronique",
        "Virginie",
        "Zoé",
    )

    first_names = first_names_male + first_names_female

    # Source:
    # https://fr.wikipedia.org/wiki/Liste_des_noms_de_famille_les_plus_courants_au_Qu%C3%A9bec
    last_names = (
        "Tremblay",
        "Gagnon",
        "Roy",
        "Côté",
        "Bouchard",
        "Gauthier",
        "Morin",
        "Lavoie",
        "Fortin",
        "Gagné",
        "Ouellet",
        "Pelletier",
        "Bélanger",
        "Lévesque",
        "Bergeron",
        "Leblanc",
        "Paquette",
        "Girard",
        "Simard",
        "Boucher",
        "Caron",
        "Beaulieu",
        "Cloutier",
        "Dubé",
        "Poirier",
        "Fournier",
        "Lapointe",
        "Leclerc",
        "Lefebvre",
        "Poulin",
        "Thibault",
        "St-Pierre",
        "Nadeau",
        "Martin",
        "Landry",
        "Martel",
        "Bédard",
        "Grenier",
        "Lessard",
        "Bernier",
        "Richard",
        "Michaud",
        "Hébert",
        "Desjardins",
        "Couture",
        "Turcotte",
        "Lachance",
        "Parent",
        "Blais",
        "Gosselin",
        "Savard",
        "Proulx",
        "Beaudoin",
        "Demers",
        "Perreault",
        "Boudreau",
        "Lemieux",
        "Cyr",
        "Perron",
        "Dufour",
        "Dion",
        "Mercier",
        "Bolduc",
        "Bérubé",
        "Boisvert",
        "Langlois",
        "Ménard",
        "Therrien",
        "Plante",
        "Bilodeau",
        "Blanchette",
        "Dubois",
        "Champagne",
        "Paradis",
        "Fortier",
        "Arsenault",
        "Dupuis",
        "Gaudreault",
        "Hamel",
        "Houle",
        "Villeneuve",
        "Rousseau",
        "Gravel",
        "Thériault",
        "Lemay",
        "Robert",
        "Allard",
        "Deschênes",
        "Giroux",
        "Guay",
        "Leduc",
        "Boivin",
        "Charbonneau",
        "Lambert",
        "Raymond",
        "Vachon",
        "Gilbert",
        "Audet",
        "Jean",
        "Larouche",
        "Legault",
        "Trudel",
        "Fontaine",
        "Picard",
        "Labelle",
        "Lacroix",
        "Jacques",
        "Moreau",
        "Carrier",
        "Bernard",
        "Desrosiers",
        "Goulet",
        "Renaud",
        "Dionne",
        "Lapierre",
        "Vaillancourt",
        "Fillion",
        "Lalonde",
        "Tessier",
        "Bertrand",
        "Tardif",
        "Lepage",
        "Gingras",
        "Benoît",
        "Rioux",
        "Giguère",
        "Drouin",
        "Harvey",
        "Lauzon",
        "Nguyen",
        "Gendron",
        "Boutin",
        "Laflamme",
        "Vallée",
        "Dumont",
        "Breton",
        "Paré",
        "Paquin",
        "Robitaille",
        "Gélinas",
        "Duchesne",
        "Lussier",
        "Séguin",
        "Veilleux",
        "Potvin",
        "Gervais",
        "Pépin",
        "Laroche",
        "Morissette",
        "Charron",
        "Lavallée",
        "Laplante",
        "Chabot",
        "Brunet",
        "Vézina",
        "Desrochers",
        "Labrecque",
        "Coulombe",
        "Tanguay",
        "Chouinard",
        "Noël",
        "Pouliot",
        "Lacasse",
        "Daigle",
        "Marcoux",
        "Lamontagne",
        "Turgeon",
        "Larocque",
        "Roberge",
        "Auger",
        "Massé",
        "Pilon",
        "Racine",
        "Dallaire",
        "Émond",
        "Grégoire",
        "Beauregard",
        "Smith",
        "Denis",
        "Lebel",
        "Blouin",
        "Martineau",
        "Labbé",
        "Beauchamp",
        "St-Onge",
        "Charette",
        "Dupont",
        "Létourneau",
        "Rodrigue",
        "Cormier",
        "Rivard",
        "Mathieu",
        "Asselin",
        "St-Jean",
        "Plourde",
        "Thibodeau",
        "Bélisle",
        "St-Laurent",
        "Godin",
        "Desbiens",
        "Lavigne",
        "Doucet",
        "Labonté",
        "Marchand",
        "Brassard",
        "Forget",
        "Patel",
        "Marcotte",
        "Béland",
        "Larose",
        "Duval",
        "Archambault",
        "Maltais",
        "Trépanier",
        "Laliberté",
        "Bisson",
        "Brisson",
        "Dufresne",
        "Beaudry",
        "Chartrand",
        "Houde",
        "Fréchette",
        "Lafontaine",
        "Guillemette",
        "Drolet",
        "Vincent",
        "Richer",
        "Germain",
        "Larivière",
        "Ferland",
        "Trottier",
        "Piché",
        "Boulanger",
        "Sirois",
        "Charest",
        "Provost",
        "Durand",
        "Dumas",
        "Soucy",
        "Lamoureux",
        "Lachapelle",
        "Bégin",
        "Boily",
        "Croteau",
        "Savoie",
        "Provencher",
        "Prévost",
        "Duguay",
        "Lemire",
        "Delisle",
    )



from __future__ import annotations

import copy
import functools
import re

from collections import OrderedDict
from random import Random
from typing import Any, Callable, Dict, List, Optional, Pattern, Sequence, Tuple, TypeVar, Union

from .config import DEFAULT_LOCALE
from .exceptions import UniquenessException
from .factory import Factory
from .generator import Generator, random
from .typing import SeedType
from .utils.distribution import choices_distribution

_UNIQUE_ATTEMPTS = 1000

RetType = TypeVar("RetType")


class Faker:
    """Proxy class capable of supporting multiple locales"""

    cache_pattern: Pattern = re.compile(r"^_cached_\w*_mapping$")
    generator_attrs = [
        attr for attr in dir(Generator) if not attr.startswith("__") and attr not in ["seed", "seed_instance", "random"]
    ]

    def __init__(
        self,
        locale: Optional[Union[str, Sequence[str], Dict[str, Union[int, float]]]] = None,
        providers: Optional[List[str]] = None,
        generator: Optional[Generator] = None,
        includes: Optional[List[str]] = None,
        use_weighting: bool = True,
        **config: Any,
    ) -> None:
        self._factory_map: OrderedDict[str, Generator | Faker] = OrderedDict()
        self._weights = None
        self._unique_proxy = UniqueProxy(self)
        self._optional_proxy = OptionalProxy(self)

        if isinstance(locale, str):
            locales = [locale.replace("-", "_")]

        # This guarantees a FIFO ordering of elements in `locales` based on the final
        # locale string while discarding duplicates after processing
        elif isinstance(locale, (list, tuple, set)):
            locales = []
            for code in locale:
                if not isinstance(code, str):
                    raise TypeError(f'The locale "{str(code)}" must be a string.')
                final_locale = code.replace("-", "_")
                if final_locale not in locales:
                    locales.append(final_locale)

        elif isinstance(locale, (OrderedDict, dict)):
            assert all(isinstance(v, (int, float)) for v in locale.values())
            odict = OrderedDict()
            for k, v in locale.items():
                key = k.replace("-", "_")
                odict[key] = v
            locales = list(odict.keys())
            self._weights = list(odict.values())

        else:
            locales = [DEFAULT_LOCALE]

        if len(locales) == 1:
            self._factory_map[locales[0]] = Factory.create(
                locales[0],
                providers,
                generator,
                includes,
                use_weighting=use_weighting,
                **config,
            )
        else:
            for locale in locales:
                self._factory_map[locale] = Faker(
                    locale,
                    providers,
                    generator,
                    includes,
                    use_weighting=use_weighting,
                    **config,
                )

        self._locales = locales
        self._factories = list(self._factory_map.values())

    def __dir__(self):
        attributes = set(super(Faker, self).__dir__())
        for factory in self.factories:
            attributes |= {attr for attr in dir(factory) if not attr.startswith("_")}
        return sorted(attributes)

    def __getitem__(self, locale: str) -> Faker:
        if locale.replace("-", "_") in self.locales and len(self.locales) == 1:
            return self
        instance = self._factory_map[locale.replace("-", "_")]
        assert isinstance(instance, Faker)  # for mypy
        return instance

    def __getattribute__(self, attr: str) -> Any:
        """
        Handles the "attribute resolution" behavior for declared members of this proxy class

        The class method `seed` cannot be called from an instance.

        :param attr: attribute name
        :return: the appropriate attribute
        """
        if attr == "seed":
            msg = "Calling `.seed()` on instances is deprecated. " "Use the class method `Faker.seed()` instead."
            raise TypeError(msg)
        else:
            return super().__getattribute__(attr)

    def __getattr__(self, attr: str) -> Any:
        """
        Handles cache access and proxying behavior

        :param attr: attribute name
        :return: the appropriate attribute
        """
        if len(self._factories) == 1:
            return getattr(self._factories[0], attr)
        elif attr in self.generator_attrs:
            msg = "Proxying calls to `%s` is not implemented in multiple locale mode." % attr
            raise NotImplementedError(msg)
        elif self.cache_pattern.match(attr):
            msg = "Cached attribute `%s` does not exist" % attr
            raise AttributeError(msg)
        else:
            factory = self._select_factory(attr)
            return getattr(factory, attr)

    def __deepcopy__(self, memodict):
        cls = self.__class__
        result = cls.__new__(cls)
        result._locales = copy.deepcopy(self._locales)
        result._factories = copy.deepcopy(self._factories)
        result._factory_map = copy.deepcopy(self._factory_map)
        result._weights = copy.deepcopy(self._weights)
        result._unique_proxy = UniqueProxy(self)
        result._unique_proxy._seen = {k: {result._unique_proxy._sentinel} for k in self._unique_proxy._seen.keys()}
        return result

    def __setstate__(self, state: Any) -> None:
        self.__dict__.update(state)

    @property
    def unique(self) -> "UniqueProxy":
        return self._unique_proxy

    @property
    def optional(self) -> "OptionalProxy":
        return self._optional_proxy

    def _select_factory(self, method_name: str) -> Factory:
        """
        Returns a random factory that supports the provider method

        :param method_name: Name of provider method
        :return: A factory that supports the provider method
        """

        factories, weights = self._map_provider_method(method_name)

        if len(factories) == 0:
            msg = f"No generator object has attribute {method_name!r}"
            raise AttributeError(msg)
        elif len(factories) == 1:
            return factories[0]

        if weights:
            factory = self._select_factory_distribution(factories, weights)
        else:
            factory = self._select_factory_choice(factories)
        return factory

    def _select_factory_distribution(self, factories, weights):
        return choices_distribution(factories, weights, random, length=1)[0]

    def _select_factory_choice(self, factories):
        return random.choice(factories)

    def _map_provider_method(self, method_name: str) -> Tuple[List[Factory], Optional[List[float]]]:
        """
        Creates a 2-tuple of factories and weights for the given provider method name

        The first element of the tuple contains a list of compatible factories.
        The second element of the tuple contains a list of distribution weights.

        :param method_name: Name of provider method
        :return: 2-tuple (factories, weights)
        """

        # Return cached mapping if it exists for given method
        attr = f"_cached_{method_name}_mapping"
        if hasattr(self, attr):
            return getattr(self, attr)

        # Create mapping if it does not exist
        if self._weights:
            value = [
                (factory, weight)
                for factory, weight in zip(self.factories, self._weights)
                if hasattr(factory, method_name)
            ]
            factories, weights = zip(*value)
            mapping = list(factories), list(weights)
        else:
            value = [factory for factory in self.factories if hasattr(factory, method_name)]  # type: ignore
            mapping = value, None  # type: ignore

        # Then cache and return results
        setattr(self, attr, mapping)
        return mapping

    @classmethod
    def seed(cls, seed: Optional[SeedType] = None) -> None:
        """
        Hashables the shared `random.Random` object across all factories

        :param seed: seed value
        """
        Generator.seed(seed)

    def seed_instance(self, seed: Optional[SeedType] = None) -> None:
        """
        Creates and seeds a new `random.Random` object for each factory

        :param seed: seed value
        """
        for factory in self._factories:
            factory.seed_instance(seed)

    def seed_locale(self, locale: str, seed: Optional[SeedType] = None) -> None:
        """
        Creates and seeds a new `random.Random` object for the factory of the specified locale

        :param locale: locale string
        :param seed: seed value
        """
        self._factory_map[locale.replace("-", "_")].seed_instance(seed)

    @property
    def random(self) -> Random:
        """
        Proxies `random` getter calls

        In single locale mode, this will be proxied to the `random` getter
        of the only internal `Generator` object. Subclasses will have to
        implement desired behavior in multiple locale mode.
        """

        if len(self._factories) == 1:
            return self._factories[0].random
        else:
            msg = "Proxying `random` getter calls is not implemented in multiple locale mode."
            raise NotImplementedError(msg)

    @random.setter
    def random(self, value: Random) -> None:
        """
        Proxies `random` setter calls

        In single locale mode, this will be proxied to the `random` setter
        of the only internal `Generator` object. Subclasses will have to
        implement desired behavior in multiple locale mode.
        """

        if len(self._factories) == 1:
            self._factories[0].random = value
        else:
            msg = "Proxying `random` setter calls is not implemented in multiple locale mode."
            raise NotImplementedError(msg)

    @property
    def locales(self) -> List[str]:
        return list(self._locales)

    @property
    def weights(self) -> Optional[List[Union[int, float]]]:
        return self._weights

    @property
    def factories(self) -> List[Generator | Faker]:
        return self._factories

    def items(self) -> List[Tuple[str, Generator | Faker]]:
        return list(self._factory_map.items())


class UniqueProxy:
    def __init__(self, proxy: Faker):
        self._proxy = proxy
        self._seen: Dict = {}
        self._sentinel = object()

    def clear(self) -> None:
        self._seen = {}

    def __getattr__(self, name: str) -> Any:
        obj = getattr(self._proxy, name)
        if callable(obj):
            return self._wrap(name, obj)
        else:
            raise TypeError("Accessing non-functions through .unique is not supported.")

    def __getstate__(self):
        # Copy the object's state from self.__dict__ which contains
        # all our instance attributes. Always use the dict.copy()
        # method to avoid modifying the original state.
        state = self.__dict__.copy()
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)

    def _wrap(self, name: str, function: Callable) -> Callable:
        @functools.wraps(function)
        def wrapper(*args, **kwargs):
            key = (name, args, tuple(sorted(kwargs.items())))

            generated = self._seen.setdefault(key, {self._sentinel})

            # With use of a sentinel value rather than None, we leave
            # None open as a valid return value.
            retval = self._sentinel

            for i in range(_UNIQUE_ATTEMPTS):
                if retval not in generated:
                    break
                retval = function(*args, **kwargs)
            else:
                raise UniquenessException(f"Got duplicated values after {_UNIQUE_ATTEMPTS:,} iterations.")

            generated.add(retval)

            return retval

        return wrapper


class OptionalProxy:
    """
    Return either a fake value or None, with a customizable probability.
    """

    def __init__(self, proxy: Faker):
        self._proxy = proxy

    def __getattr__(self, name: str) -> Any:
        obj = getattr(self._proxy, name)
        if callable(obj):
            return self._wrap(name, obj)
        else:
            raise TypeError("Accessing non-functions through .optional is not supported.")

    def __getstate__(self):
        # Copy the object's state from self.__dict__ which contains
        # all our instance attributes. Always use the dict.copy()
        # method to avoid modifying the original state.
        state = self.__dict__.copy()
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)

    def _wrap(self, name: str, function: Callable[..., RetType]) -> Callable[..., Optional[RetType]]:
        @functools.wraps(function)
        def wrapper(*args: Any, prob: float = 0.5, **kwargs: Any) -> Optional[RetType]:
            if not 0 < prob <= 1.0:
                raise ValueError("prob must be between 0 and 1")
            return function(*args, **kwargs) if self._proxy.boolean(chance_of_getting_true=int(prob * 100)) else None

        return wrapper



from .. import Provider as BaseProvider


class Provider(BaseProvider):
    """
    source: https://ja.wikipedia.org/wiki/%E8%81%B7%E6%A5%AD%E4%B8%80%E8%A6%A7
    """

    jobs = [
        "アイドル",
        "アーティスト",
        "アートディレクター",
        "アナウンサー",
        "アニメーター",
        "医師",
        "イラストレーター",
        "医療事務員",
        "ウェディングプランナー",
        "ウェブデザイナー",
        "占い師",
        "運転士",
        "映画監督",
        "営業",
        "栄養士",
        "エステティシャン",
        "絵本作家",
        "演歌歌手",
        "エンジニア" "演奏家",
        "お笑い芸人",
        "音楽家",
        "音響技術者",
        "介護ヘルパー",
        "気象予報士",
        "脚本家",
        "救急救命士",
        "行政書士",
        "グラフィックデザイナー",
        "経営者",
        "検察官",
        "ゲームクリエイター",
        "建築家",
        "航海士",
        "コピーライター",
        "高等学校教員",
        "公認会計士",
        "公務員",
        "裁判官",
        "作曲家",
        "歯科医師",
        "司法書士",
        "小説家",
        "寿司職人",
        "測量士",
        "大学教授",
        "調理師",
        "電気工事士",
        "農家",
        "配管工",
        "バスガイド",
        "花火師",
        "漫画家",
        "モデル",
        "薬剤師",
        "YouTuber",
        "和紙職人",
    ]
